from __future__ import annotations

from pathlib import Path
from typing import Any, Generic, TypeVar

import pydantic
import yaml

from ..exceptions import BadConfig


class CustomBaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid", validate_default=True)


T = TypeVar("T", bound=pydantic.BaseModel)


class YamlLoaderMixin(Generic[T]):
    @classmethod
    def from_yaml(cls: type[T], path: Path) -> T:  # type: ignore[misc]
        try:
            with path.open() as f:
                return cls(**yaml.safe_load(f))
        except FileNotFoundError:
            raise BadConfig(f"File {path} not found")
        except TypeError as e:
            raise BadConfig(f"Config YAML error:\n{e}")
        except yaml.YAMLError as e:
            raise BadConfig(f"Config YAML error:\n{e}")
        except pydantic.ValidationError as e:
            raise BadConfig(f"Config Validation error:\n{e}")

    def to_yaml(self: T, path: Path) -> None:  # type: ignore[misc]
        with path.open("w") as f:
            yaml.dump(self.model_dump(), f)

    @classmethod
    def get_json_schema(cls: type[T]) -> dict[str, Any]:  # type: ignore[misc]
        return cls.model_json_schema()



from __future__ import annotations

from typing import Literal, Union

from ..exceptions import PluginExecutionFailed
from .base import PluginABC, PluginOutput


class AggregatePlugin(PluginABC):
    """Given scores and optional weights and strategy, aggregate them, return the score."""

    name = "aggregate"

    class Args(PluginABC.Args):
        scores: list[float]
        weights: Union[list[float], None] = None  # as pydantic does not support | in older python versions
        strategy: Literal["mean", "sum", "min", "max", "product"] = "mean"
        # TODO: validate for weights: len weights should be equal to len scores
        # TODO: validate not empty scores

    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:  # type: ignore[override]
        weights = args.weights or ([1.0] * len(args.scores))

        if len(args.scores) != len(weights):
            raise PluginExecutionFailed(
                f"Length of scores ({len(args.scores)}) and weights ({len(weights)}) does not match",
                output=f"Length of scores ({len(args.scores)}) and weights ({len(weights)}) does not match",
            )

        if len(args.scores) == 0 or len(weights) == 0:
            raise PluginExecutionFailed(
                f"Length of scores ({len(args.scores)}) or weights ({len(weights)}) is zero",
                output=f"Length of scores ({len(args.scores)}) or weights ({len(weights)}) is zero",
            )

        weighted_scores = [score * weight for score, weight in zip(args.scores, weights)]

        if args.strategy == "mean":
            score = sum(weighted_scores) / len(weighted_scores)
        elif args.strategy == "sum":
            score = sum(weighted_scores)
        elif args.strategy == "min":
            score = min(weighted_scores)
        elif args.strategy == "max":
            score = max(weighted_scores)
        elif args.strategy == "product":
            from functools import reduce

            score = reduce(lambda x, y: x * y, weighted_scores)
        else:  # pragma: no cover
            assert False, "Not reachable"

        return PluginOutput(
            output=(
                f"Get scores:  {args.scores}\n"
                f"Get weights: {args.weights}\n"
                f"Aggregate weighted scores {weighted_scores} with strategy {args.strategy}\n"
                f"Score: {score:.2f}"
            ),
            percentage=score,
        )



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

from ..exceptions import (
    BuildFailedError,
    ExecutionFailedError,
    StylecheckFailedError,
    TestsFailedError,
    TimeoutExpiredError,
)
from ..utils.files import check_files_contains_regexp, copy_files
from ..utils.print import print_info
from .tester import Tester


class CppTester(Tester):

    @dataclass
    class TaskTestConfig(Tester.TaskTestConfig):
        allow_change: list[str] = field(default_factory=list)
        forbidden_regexp: list[str] = field(default_factory=list)
        copy_to_build: list[str] = field(default_factory=list)

        linter: bool = True
        build_type: str = 'Asan'
        is_crash_me: bool = False

        tests: list[str] = field(default_factory=list)
        input_file: dict[str, str] = field(default_factory=dict)
        args: dict[str, list[str]] = field(default_factory=dict)
        timeout: float = 60.
        capture_output: bool = True

        def __post_init__(
                self,
        ) -> None:
            assert self.tests
            assert self.allow_change

    def _gen_build(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            source_dir: Path,
            public_tests_dir: Path | None,
            private_tests_dir: Path | None,
            tests_root_dir: Path,
            sandbox: bool = True,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> None:
        check_files_contains_regexp(
            source_dir,
            regexps=test_config.forbidden_regexp,
            patterns=test_config.allow_change,
            raise_on_found=True,
        )
        task_dir = public_tests_dir
        self._executor(
            copy_files,
            source=source_dir,
            target=task_dir,
            patterns=test_config.allow_change,
            verbose=verbose,
        )
        self._executor(
            copy_files,
            source=task_dir,
            target=build_dir,
            patterns=test_config.copy_to_build,
            verbose=verbose,
        )

        try:
            print_info('Running cmake...', color='orange')
            self._executor(
                ['cmake', '-G', 'Ninja', str(tests_root_dir),
                 '-DGRADER=YES', '-DENABLE_PRIVATE_TESTS=YES',
                 f'-DCMAKE_BUILD_TYPE={test_config.build_type}'],
                cwd=build_dir,
                verbose=verbose,
            )
        except ExecutionFailedError:
            print_info('ERROR', color='red')
            raise BuildFailedError('cmake execution failed')

        for test_binary in test_config.tests:
            try:
                print_info(f'Building {test_binary}...', color='orange')
                self._executor(
                    ['ninja', '-v', test_binary],
                    cwd=build_dir,
                    verbose=verbose,
                )
            except ExecutionFailedError:
                print_info('ERROR', color='red')
                raise BuildFailedError(f'Can\'t build {test_binary}')

        if not test_config.linter:
            return

        try:
            print_info('Running clang format...', color='orange')
            format_path = tests_root_dir / 'run-clang-format.py'
            self._executor(
                [str(format_path), '-r', str(task_dir)],
                cwd=build_dir,
                verbose=verbose,
            )
            print_info('[No issues]')
            print_info('OK', color='green')
        except ExecutionFailedError:
            print_info('ERROR', color='red')
            raise StylecheckFailedError('Style error (clang format)')

        try:
            print_info('Running clang tidy...', color='orange')
            files = [str(file) for file in task_dir.rglob('*.cpp')]  # type: ignore
            self._executor(
                ['clang-tidy', '-p', '.', *files],
                cwd=build_dir,
                verbose=verbose,
            )
            print_info('[No issues]')
            print_info('OK', color='green')
        except ExecutionFailedError:
            print_info('ERROR', color='red')
            raise StylecheckFailedError('Style error (clang tidy)')

    def _clean_build(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            verbose: bool = False,
    ) -> None:
        self._executor(
            ['rm', '-rf', str(build_dir)],
            check=False,
            verbose=verbose,
        )

    def _run_tests(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            sandbox: bool = False,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> float:
        for test_binary in test_config.tests:
            stdin = None
            try:
                print_info(f'Running {test_binary}...', color='orange')
                args = test_config.args.get(test_binary, [])
                if test_binary in test_config.input_file:
                    stdin = open(build_dir / test_config.input_file[test_binary], 'r')
                self._executor(
                    [str(build_dir / test_binary), *args],
                    sandbox=True,
                    cwd=build_dir,
                    verbose=verbose,
                    capture_output=test_config.capture_output,
                    timeout=test_config.timeout,
                    stdin=stdin
                )
                if test_config.is_crash_me:
                    print_info('ERROR', color='red')
                    raise TestsFailedError('Program has not crashed')
                print_info('OK', color='green')
            except TimeoutExpiredError:
                print_info('ERROR', color='red')
                message = f'Your solution exceeded time limit: {test_config.timeout} seconds'
                raise TestsFailedError(message)
            except ExecutionFailedError:
                if not test_config.is_crash_me:
                    print_info('ERROR', color='red')
                    raise TestsFailedError("Test failed (wrong answer or sanitizer error)")
            finally:
                if stdin is not None:
                    stdin.close()
        if test_config.is_crash_me:
            print_info('Program has crashed', color='green')
        else:
            print_info('All tests passed', color='green')
        return 1.



"""Main executable file. Refer to cli module"""
from __future__ import annotations

import json
import os
import shutil
import sys
import tempfile
from pathlib import Path
from typing import Any

import click

from .actions.check import pre_release_check_tasks

# from .actions.contributing import create_public_mr  # type: ignore
from .actions.export import export_public_files
from .actions.grade import grade_on_ci
from .actions.grade_mr import grade_student_mrs, grade_students_mrs_to_master
from .course import CourseConfig, CourseSchedule, Task
from .course.driver import CourseDriver
from .testers import Tester
from .utils.glab import GitlabConnection
from .utils.print import print_info


ClickTypeReadableFile = click.Path(exists=True, file_okay=True, readable=True, path_type=Path)
ClickTypeReadableDirectory = click.Path(exists=True, file_okay=False, readable=True, path_type=Path)
ClickTypeWritableDirectory = click.Path(file_okay=False, writable=True, path_type=Path)


@click.group()
@click.option('-c', '--config', envvar='CHECKER_CONFIG', type=ClickTypeReadableFile, default=None,
              help='Course config path')
@click.version_option(package_name='manytask-checker')
@click.pass_context
def main(
        ctx: click.Context,
        config: Path | None,
) -> None:
    """Students' solutions *checker*"""
    # Read course config and pass it to any command
    # If not provided - read .course.yml from the root
    config = config or Path() / '.course.yml'
    if not config.exists():
        config = Path() / 'tests' / '.course.yml'
    if not config.exists():
        config = Path() / 'tools' / '.course.yml'

    if not config.exists():
        raise FileNotFoundError('Unable to find `.course.yml` config')

    execution_folder = Path()

    ctx.obj = {
        'course_config': CourseConfig.from_yaml(config),
        'execution_folder': execution_folder,
    }


@main.command()
@click.argument('root', required=False, type=ClickTypeReadableDirectory)
@click.option('--task', type=str, multiple=True, help='Task name to check')
@click.option('--group', type=str, multiple=True, help='Group name to check')
@click.option('--no-clean', is_flag=True, help='Clean or not check tmp folders')
@click.option('--dry-run', is_flag=True, help='Do not execute anything, only print')
@click.option('--parallelize', is_flag=True, help='Execute parallel checking of tasks')
@click.option('--num-processes', type=int, default=None, help='Num of processes parallel checking (default: unlimited)')
@click.option('--contributing', is_flag=True, help='Run task check for students` contribution (decrease verbosity)')
@click.pass_context
def check(
        ctx: click.Context,
        root: Path | None = None,
        task: list[str] | None = None,
        group: list[str] | None = None,
        no_clean: bool = False,
        dry_run: bool = False,
        parallelize: bool = False,
        num_processes: int | None = None,
        contributing: bool = False,
) -> None:
    """Run task pre-release checking"""
    context: dict[str, Any] = ctx.obj
    course_config: CourseConfig = context['course_config']
    execution_folder: Path = context['execution_folder']

    root = root or execution_folder
    private_course_driver = CourseDriver(
        root_dir=root,
        repo_type='private',
        layout=course_config.layout,
    )
    course_schedule = CourseSchedule(
        deadlines_config=private_course_driver.get_deadlines_file_path(),
    )
    tester = Tester.create(
        root=root,
        course_config=course_config,
        cleanup=not no_clean,
        dry_run=dry_run,
    )

    tasks: list[Task] | None = None
    if group:
        tasks = []
        for group_name in group:
            if group_name in course_schedule.groups:
                tasks.extend(course_schedule.groups[group_name].tasks)
            else:
                print_info(f'Provided wrong group name: {group_name}', color='red')
                sys.exit(1)
    elif task:
        tasks = []
        for task_name in task:
            if task_name in course_schedule.tasks:
                tasks.append(course_schedule.tasks[task_name])
            else:
                print_info(f'Provided wrong task name: {task_name}', color='red')
                sys.exit(1)

    pre_release_check_tasks(
        course_schedule,
        private_course_driver,
        tester,
        tasks=tasks,
        parallelize=parallelize,
        num_processes=num_processes,
        contributing=contributing,
    )


@main.command()
@click.argument('reference_root', required=False, type=ClickTypeReadableDirectory)
@click.option('--test-full-groups', is_flag=True, help='Test all tasks in changed groups')
@click.option('--task', type=str, multiple=True, help='Task name to check')
@click.option('--group', type=str, multiple=True, help='Group name to check')
@click.pass_context
def grade(
        ctx: click.Context,
        reference_root: Path | None = None,
        test_full_groups: bool = False,
        task: list[str] | None = None,
        group: list[str] | None = None,
) -> None:
    """Run student's tasks (current ci user)"""
    context: dict[str, Any] = ctx.obj
    course_config: CourseConfig = context['course_config']
    execution_folder: Path = context['execution_folder']

    reference_root = reference_root or execution_folder
    public_course_driver = CourseDriver(
        root_dir=Path(os.environ['CI_PROJECT_DIR']),
        repo_type='public',
        layout=course_config.layout,
    )
    private_course_driver = CourseDriver(
        root_dir=reference_root,
        repo_type='private',
        layout=course_config.layout,
    )
    course_schedule = CourseSchedule(
        deadlines_config=private_course_driver.get_deadlines_file_path(),
    )

    tester = Tester.create(
        root=execution_folder,
        course_config=course_config,
    )

    grade_on_ci(
        course_config,
        course_schedule,
        public_course_driver,
        private_course_driver,
        tester,
        test_full_groups=test_full_groups,
        task=task,
        group=group
    )
    # TODO: think inspect


@main.command()
@click.argument('reference_root', required=False, type=ClickTypeReadableDirectory)
@click.option('--dry-run', is_flag=True, help='Do not execute anything, only print')
@click.pass_context
def grade_mrs(
        ctx: click.Context,
        reference_root: Path | None = None,
        dry_run: bool = False,
) -> None:
    """Run student's MRs grading (current git user)"""
    context: dict[str, Any] = ctx.obj
    course_config: CourseConfig = context['course_config']
    execution_folder: Path = context['execution_folder']

    reference_root = reference_root or execution_folder
    public_course_driver = CourseDriver(
        root_dir=Path(os.environ['CI_PROJECT_DIR']),
        repo_type='public',
        layout=course_config.layout,
    )
    private_course_driver = CourseDriver(
        root_dir=reference_root,
        repo_type='private',
        layout=course_config.layout,
    )
    course_schedule = CourseSchedule(
        deadlines_config=private_course_driver.get_deadlines_file_path(),
    )

    username = os.environ['CI_PROJECT_NAME']

    gitlab_connection = GitlabConnection(
        gitlab_host_url=course_config.gitlab_url,
        job_token=os.environ.get('CI_JOB_TOKEN'),
    )

    grade_student_mrs(
        course_config,
        course_schedule,
        public_course_driver,
        gitlab_connection,
        username,
        dry_run=dry_run,
    )
    # TODO: think inspect


@main.command()
@click.argument('root', required=False, type=ClickTypeReadableDirectory)
@click.option('--dry-run', is_flag=True, help='Do not execute anything, only print')
@click.pass_context
def grade_students_mrs(
        ctx: click.Context,
        root: Path | None = None,
        dry_run: bool = False,
) -> None:
    """Run students' MRs grading (all users)"""
    context: dict[str, Any] = ctx.obj
    course_config: CourseConfig = context['course_config']
    execution_folder: Path = context['execution_folder']

    root = root or execution_folder
    private_course_driver = CourseDriver(
        root_dir=root,
        repo_type='private',
        layout=course_config.layout,
    )
    course_schedule = CourseSchedule(
        deadlines_config=private_course_driver.get_deadlines_file_path(),
    )

    gitlab_connection = GitlabConnection(
        gitlab_host_url=course_config.gitlab_url,
        private_token=course_config.gitlab_service_token,
    )

    grade_students_mrs_to_master(
        course_config,
        course_schedule,
        private_course_driver,
        gitlab_connection,
        dry_run=dry_run,
    )


@main.command()
@click.argument('root', required=False, type=ClickTypeReadableDirectory)
@click.option('--export-dir', type=ClickTypeWritableDirectory, help='TEMP dir to export into')
@click.option('--dry-run', is_flag=True, help='Do not execute anything, only print')
@click.option('--no-cleanup', is_flag=True, help='Do not cleanup export dir')
@click.pass_context
def export_public(
        ctx: click.Context,
        root: Path | None = None,
        export_dir: Path | None = None,
        dry_run: bool = False,
        no_cleanup: bool = False,
) -> None:
    """Export enabled tasks to public repo"""
    context: dict[str, Any] = ctx.obj
    course_config: CourseConfig = context['course_config']
    execution_folder: Path = context['execution_folder']

    root = root or execution_folder

    export_dir = export_dir or Path(tempfile.mkdtemp())
    if not export_dir.exists():
        export_dir.mkdir(exist_ok=True, parents=True)

    public_course_driver = CourseDriver(
        root_dir=export_dir,
        repo_type='public',
        layout=course_config.layout,
    )
    private_course_driver = CourseDriver(
        root_dir=root,
        repo_type='private',
        layout=course_config.layout,
    )
    course_schedule = CourseSchedule(
        deadlines_config=private_course_driver.get_deadlines_file_path(),
    )

    export_public_files(
        course_config,
        course_schedule,
        public_course_driver,
        private_course_driver,
        export_dir,
        dry_run=dry_run,
    )

    if no_cleanup:
        print_info(f'No cleanup flag. Exported files stored in {export_dir}.')
    else:
        print_info(f'Cleanup flag. Export dir {export_dir} removed.')
        shutil.rmtree(export_dir)


# @main.command()
# @click.option('--dry-run', is_flag=True, help='Do not execute anything, only print')
# @click.pass_context
def create_contributing_mr(
        ctx: click.Context,
        dry_run: bool = False,
) -> None:
    """Move public project to private as MR"""
    context: dict[str, Any] = ctx.obj
    course_config: CourseConfig = context['course_config']
    # execution_folder: Path = context['execution_folder']

    trigger_payload = os.environ.get('TRIGGER_PAYLOAD', 'None')
    print_info('trigger_payload', trigger_payload)

    # trigger_payload_dict = json.loads(trigger_payload)
    with open(trigger_payload, 'r') as json_file:
        trigger_payload_dict = json.load(json_file)

    event_type = trigger_payload_dict['event_type']

    if event_type != 'merge_request':
        print_info(f'event_type = {event_type}. Skip it.', color='orange')
        return

    object_attributes = trigger_payload_dict['object_attributes']
    merge_commit_sha = object_attributes['merge_commit_sha']

    if merge_commit_sha is None:
        print_info('merge_commit_sha = None. Skip it.', color='orange')
        return

    mr_state = object_attributes['state']
    target_branch = object_attributes['target_branch']

    if mr_state != 'merged':
        print_info(f'mr_state = {mr_state}. Skip it.', color='orange')
        return

    if target_branch != course_config.default_branch:
        print_info(f'target_branch = {target_branch}. Skip it.', color='orange')
        return

    # create_public_mr(course_config, object_attributes, dry_run=dry_run)


if __name__ == '__main__':  # pragma: nocover
    main()



"""
Classes to map course schedule to real filesystem
"""
from __future__ import annotations

import os
from pathlib import Path
from warnings import warn

from ..exceptions import BadConfig
from ..utils import print_info
from .schedule import Group, Task


class CourseDriver:
    """The interlayer between course and file system
    Course can have different layouts;
    You can select 2 layouts: for private and for public repo
    for script to know how to read private script and how to write to the public repo
    Now implemented: @see self.PUBLIC_LAYOUTS, self.PRIVATE_LAYOUTS

    * flat [deprecated] (public & private)
        - .gitignore
        - .gitlab-ci.yml
        - .releaser-ci.yml
        - README.md
        - task_1/
        - ...
        - tests/
            - .course.yml
            - .deadlines.yml
            - task_1/
            - ...

    * groups (public & private)
        - .course.yml
        - .deadlines.yml
        - .gitignore
        - .gitlab-ci.yml
        - .releaser-ci.yml
        - README.md
        - group_1/
            - task_1/
            - ...
        - ...
        - lectures/
            - group_1/
            - ...
        - solutions/
            - group_1/
            - ...
        - tests/
            - group_1/
                - task_1/
                - ...

    * lectures (private)
        - .course.yml
        - .deadlines.yml
        - .gitignore
        - .gitlab-ci.yml
        - .releaser-ci.yml
        - README.md
        - group_1/
            lecture/ [optional]
            review/ [optional]
            tasks/
                - task_1/
                    - private/ [optional]
                        - test_private.py
                    - public/ [optional]
                        - test_public.py
                    - template/ [optional]
                        - solution.py
                    - solution/
                        - solution.py
                    - README.md
                    - .tester.json [optional]
                - ...
        - ...
    * lectures (public)
        - ...
        - group_1/
            lecture/ [optional]
            review/ [optional]
            tasks/
                - task_1/
                    - test_public.py
                    - solution.py
                    - README.md
                    - .tester.json [optional]
                - ...
        - ...

    For templates:
        * search - will search template in public folder
        * create - will search gold solution in private folder and create template from it
        * create_or_search - will search template in public folder or will create template from gold solution
    """

    LAYOUTS = ['flat', 'groups', 'lectures']
    TEMPLATES = ['create', 'search', 'create_or_search']
    REPO_TYPES = ['public', 'private']

    def __init__(
            self,
            root_dir: Path,
            repo_type: str = 'public',
            layout: str = 'groups',
            template: str = 'search',
    ):
        """
        @param root_dir: Root folder of the repo to be a driver on
        @param repo_type: Type of repository public (students repos / public) or private (main private repo)
        @param layout: @see available LAYOUTS in class docstring
        @param template: @see available TEMPLATES in class var and utils -> clear_gold_solution function
        """

        assert root_dir.exists(), f'Root dir <{root_dir}> not exists'
        self.root_dir = root_dir

        assert repo_type in CourseDriver.REPO_TYPES, f'Repo type <{repo_type}> not in private, public'
        self.repo_type = repo_type

        assert layout in CourseDriver.LAYOUTS, f'Course layout <{layout}> are not implemented'
        if layout == 'flat':
            warn(f'<{layout}> layout is deprecated', DeprecationWarning)
        self.layout = layout

        assert template in CourseDriver.TEMPLATES, f'Template <{layout}> are not implemented'
        self.template = template

    def get_deadlines_file_path(
            self,
            raise_if_not_exists: bool = True,
    ) -> Path:
        if self.repo_type == 'public':
            raise BadConfig('Unable to find `deadlines` file in public repo')

        deadlines_file_path: Path
        if self.layout == 'lectures':
            deadlines_file_path = self.root_dir / '.deadlines.yml'
        elif self.layout == 'groups':
            deadlines_file_path = self.root_dir / '.deadlines.yml'
        elif self.layout == 'flat':
            deadlines_file_path = self.root_dir / 'tests' / '.deadlines.yml'
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if raise_if_not_exists and (not deadlines_file_path or not deadlines_file_path.exists()):
            raise BadConfig(f'Deadlines file <{deadlines_file_path}> not exists')

        return deadlines_file_path

    def get_group_lecture_dir(
            self,
            group: Group,
            check_exists: bool = True,
    ) -> Path | None:
        lecture_dir: Path | None = None

        if self.layout == 'lectures':
            lecture_dir = self.root_dir / group.name / 'lecture'
        elif self.layout == 'groups':
            lecture_dir = self.root_dir / 'lectures' / group.name
        elif self.layout == 'flat':
            lecture_dir = None
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if check_exists and lecture_dir and not lecture_dir.exists():
            print_info(f'Lecture dir <{lecture_dir}> not exists, set to None.')
            lecture_dir = None

        return lecture_dir

    def get_group_submissions_review_dir(
            self,
            group: Group,
            check_exists: bool = True,
    ) -> Path | None:
        review_dir: Path | None = None

        if self.layout == 'lectures':
            # both public and private
            review_dir = self.root_dir / group.name / 'review'
        elif self.layout == 'groups':
            # both public and private
            review_dir = self.root_dir / 'solutions' / group.name
        elif self.layout == 'flat':
            review_dir = None
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if check_exists and review_dir and not review_dir.exists():
            print_info(f'Review dir <{review_dir}> not exists, set to None.')
            review_dir = None

        return review_dir

    def get_group_dir(
            self,
            group: Group,
            check_exists: bool = True,
    ) -> Path | None:
        group_root_dir: Path | None = None

        if self.layout == 'lectures':
            group_root_dir = self.root_dir / group.name
        elif self.layout == 'groups':
            group_root_dir = self.root_dir / group.name
        elif self.layout == 'flat':
            group_root_dir = None
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if check_exists and group_root_dir and not group_root_dir.exists():
            print_info(f'Group dir <{group_root_dir}> not exists, set to None.')
            group_root_dir = None

        return group_root_dir

    def get_task_dir(
            self,
            task: Task,
            check_exists: bool = True,
    ) -> Path | None:
        task_root_dir: Path | None = None

        if self.layout == 'lectures':
            task_root_dir = self.root_dir / task.group.name / 'tasks' / task.name
        elif self.layout == 'groups':
            task_root_dir = self.root_dir / task.group.name / task.name
        elif self.layout == 'flat':
            task_root_dir = self.root_dir / task.name
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if check_exists and task_root_dir and not task_root_dir.exists():
            print_info(f'Task dir <{task_root_dir}> not exists, set to None.')
            task_root_dir = None

        return task_root_dir

    def get_task_solution_dir(
            self,
            task: Task,
            check_exists: bool = True,
    ) -> Path | None:
        task_solution_dir: Path | None = None

        if self.layout == 'lectures':
            if self.repo_type == 'private':
                task_solution_dir = self.root_dir / task.group.name / 'tasks' / task.name / 'solution'
            else:
                task_solution_dir = self.root_dir / task.group.name / 'tasks' / task.name
        elif self.layout == 'groups':
            if self.repo_type == 'private':
                task_solution_dir = self.root_dir / 'tests' / task.group.name / task.name
            else:
                task_solution_dir = self.root_dir / task.group.name / task.name
        elif self.layout == 'flat':
            if self.repo_type == 'private':
                task_solution_dir = self.root_dir / 'tests' / task.name
            else:
                task_solution_dir = self.root_dir / task.name
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if check_exists and task_solution_dir and not task_solution_dir.exists():
            print_info(f'Task solution dir <{task_solution_dir}> not exists, set to None.')
            task_solution_dir = None

        return task_solution_dir

    def get_task_template_dir(
            self,
            task: Task,
            check_exists: bool = True,
    ) -> Path | None:
        task_template_dir: Path | None = None

        if self.layout == 'lectures':
            if self.repo_type == 'private':
                task_template_dir = self.root_dir / task.group.name / 'tasks' / task.name / 'template'
            else:
                task_template_dir = self.root_dir / task.group.name / 'tasks' / task.name
        elif self.layout == 'groups':
            # both public and private
            task_template_dir = self.root_dir / task.group.name / task.name
        elif self.layout == 'flat':
            # both public and private
            task_template_dir = self.root_dir / task.name
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if check_exists and task_template_dir and not task_template_dir.exists():
            print_info(f'Task template dir <{task_template_dir}> not exists, set to None.')
            task_template_dir = None

        return task_template_dir

    def get_task_public_test_dir(
            self,
            task: Task,
            check_exists: bool = True,
    ) -> Path | None:
        public_tests_dir: Path | None = None

        if self.layout == 'lectures':
            if self.repo_type == 'private':
                public_tests_dir = self.root_dir / task.group.name / 'tasks' / task.name / 'public'
            else:
                public_tests_dir = self.root_dir / task.group.name / 'tasks' / task.name
        elif self.layout == 'groups':
            # both public and private
            public_tests_dir = self.root_dir / task.group.name / task.name
        elif self.layout == 'flat':
            # both public and private
            public_tests_dir = self.root_dir / task.name
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if check_exists and public_tests_dir and not public_tests_dir.exists():
            print_info(f'Task public tests dir <{public_tests_dir}> not exists, set to None.')
            public_tests_dir = None

        return public_tests_dir

    def get_task_private_test_dir(
            self,
            task: Task,
            check_exists: bool = True,
    ) -> Path | None:
        private_tests_dir: Path | None = None

        if self.layout == 'lectures':
            if self.repo_type == 'private':
                private_tests_dir = self.root_dir / task.group.name / 'tasks' / task.name / 'private'
            else:
                private_tests_dir = None
        elif self.layout == 'groups':
            if self.repo_type == 'private':
                private_tests_dir = self.root_dir / 'tests' / task.group.name / task.name
            else:
                private_tests_dir = None
        elif self.layout == 'flat':
            if self.repo_type == 'private':
                private_tests_dir = self.root_dir / 'tests' / task.name
            else:
                private_tests_dir = None
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if check_exists and private_tests_dir and not private_tests_dir.exists():
            print_info(f'Task private tests dir <{private_tests_dir}> not exists, set to None.')
            private_tests_dir = None

        return private_tests_dir

    def get_task_config_dir(
            self,
            task: Task,
            check_exists: bool = True,
    ) -> Path | None:
        config_dir: Path | None = None

        if self.layout == 'lectures':
            if self.repo_type == 'private':
                config_dir = self.root_dir / task.group.name / 'tasks' / task.name
            else:
                config_dir = None
        elif self.layout == 'groups':
            if self.repo_type == 'private':
                config_dir = self.root_dir / 'tests' / task.group.name / task.name
            else:
                config_dir = None
        elif self.layout == 'flat':
            if self.repo_type == 'private':
                config_dir = self.root_dir / 'tests' / task.name
            else:
                config_dir = None
        else:
            assert False, 'Not Reachable'  # pragma: no cover

        if check_exists and config_dir and not config_dir.exists():
            print_info(f'Task config dir <{config_dir}> not exists, set to None.')
            config_dir = None

        return config_dir

    def get_task_dir_name(
            self,
            path: str,
    ) -> str | None:
        path_split = path.split(os.path.sep, maxsplit=3)
        if len(path_split) < 2:  # Changed file not in subdir
            return None
        if self.layout == 'lectures':
            if len(path_split) < 3:
                return None
            return path_split[2]
        elif self.layout == 'groups':
            return path_split[1]
        elif self.layout == 'flat':
            return path_split[0]
        else:
            assert False, 'Not Reachable'  # pragma: no cover



from __future__ import annotations

import shutil
from pathlib import Path

from ..course import CourseConfig, CourseDriver
from ..course.schedule import CourseSchedule
from ..utils.files import filename_match_patterns
from ..utils.git import commit_push_all_repo, setup_repo_in_dir
from ..utils.print import print_info


EXPORT_IGNORE_COMMON_FILE_PATTERNS = [
    '.git', '*.docker', '.releaser-ci.yml', '.deadlines.yml', '.course.yml', '.DS_Store', '.venv',
    '.*_cache', '.github', '*.drawio', 'LICENSE', 'pyproject.toml', 'requirements.txt', 'update.sh',
    '.dockerignore', '.checker.yml', '.gitlab-ci.private.yml', '.manytask.yml', '.pre-commit-config.yaml'
]


def _get_enabled_files_and_dirs_private_to_public(
        course_config: CourseConfig,
        course_schedule: CourseSchedule,
        public_course_driver: CourseDriver,
        private_course_driver: CourseDriver,
) -> dict[Path, Path]:
    # Common staff; files only, all from private repo except ignored
    common_files: dict[Path, Path] = {
        i: public_course_driver.root_dir / i.name
        for i in private_course_driver.root_dir.glob('*')
        if i.is_file() and not filename_match_patterns(i, EXPORT_IGNORE_COMMON_FILE_PATTERNS)
    }

    # Course docs
    course_docs: dict[Path, Path] = dict()
    if (private_course_driver.root_dir / 'docs').exists():
        course_docs.update({
            private_course_driver.root_dir / 'docs': public_course_driver.root_dir / 'docs',
        })
    if (private_course_driver.root_dir / 'images').exists():
        course_docs.update({
            private_course_driver.root_dir / 'images': public_course_driver.root_dir / 'images',
        })

    # Course tools
    course_tools: dict[Path, Path] = dict()
    if (private_course_driver.root_dir / 'tools').exists():
        course_tools = {
            i: public_course_driver.root_dir / 'tools' / i.name
            for i in (private_course_driver.root_dir / 'tools').glob('*')
            if i.is_dir() or (i.is_file() and not filename_match_patterns(i, EXPORT_IGNORE_COMMON_FILE_PATTERNS))
        }

    # Started tasks: copy template to public repo
    started_tasks_templates_dirs: dict[Path, Path] = {
        private_template_dir:
            public_course_driver.get_task_solution_dir(task, check_exists=False)  # type: ignore
        for task in course_schedule.get_tasks(enabled=True, started=True)
        if (private_template_dir := private_course_driver.get_task_template_dir(task, check_exists=True))
    }
    started_tasks_public_tests_dirs: dict[Path, Path] = {
        private_public_tests_dir:
            public_course_driver.get_task_public_test_dir(task, check_exists=False)  # type: ignore
        for task in course_schedule.get_tasks(enabled=True, started=True)
        if (private_public_tests_dir := private_course_driver.get_task_public_test_dir(task, check_exists=True))
    }
    started_tasks_common_files: dict[Path, Path] = {
        i: public_course_driver.get_task_dir(task, check_exists=False) / i.name  # type: ignore
        for task in course_schedule.get_tasks(enabled=True, started=True)
        if (private_task_dir := private_course_driver.get_task_dir(task))
        for i in private_task_dir.glob('*.*')
    }

    # Lectures for enabled groups (if any)
    started_lectures_dirs: dict[Path, Path] = {
        private_lecture_dir: public_course_driver.get_group_lecture_dir(group, check_exists=False)  # type: ignore
        for group in course_schedule.get_groups(enabled=True, started=True)
        if (private_lecture_dir := private_course_driver.get_group_lecture_dir(group, check_exists=True))
    }

    # Reviews for ended groups (if any)
    ended_reviews_dirs: dict[Path, Path] = {
        private_review_dir:
            public_course_driver.get_group_submissions_review_dir(group, check_exists=False)  # type: ignore
        for group in course_schedule.get_groups(enabled=True, ended=True)
        if (private_review_dir := private_course_driver.get_group_submissions_review_dir(group, check_exists=True))
    }

    return {
        **common_files,
        **course_docs,
        **course_tools,
        **started_tasks_templates_dirs,
        **started_tasks_public_tests_dirs,
        **started_tasks_common_files,
        **started_lectures_dirs,
        **ended_reviews_dirs,
    }


def _dirs_to_files(files_and_dirs: set[Path]) -> set[Path]:
    # Recursive add all files if we have dirs
    all_files_dirs = set()
    for i in files_and_dirs:
        if i.is_file():
            all_files_dirs.add(i)
        else:
            all_files_dirs.update(i.glob('**/*'))

    return all_files_dirs  # - set(PUBLIC_DIR.glob('.git/**/*'))


def _get_disabled_files(
        enabled_files: set[Path],
        course_driver: CourseDriver,
) -> set[Path]:
    all_files = [
        i for i in course_driver.root_dir.glob('**/*') if i.is_file()
    ]

    return set(all_files) - enabled_files - set(course_driver.root_dir.glob('.git/**/*')) - {course_driver.root_dir}


def export_public_files(
        course_config: CourseConfig,
        course_schedule: CourseSchedule,
        public_course_driver: CourseDriver,
        private_course_driver: CourseDriver,
        export_dir: Path,
        *,
        dry_run: bool = False,
) -> None:
    export_dir.mkdir(exist_ok=True, parents=True)

    if dry_run:
        print_info('Dry run. No repo setup, only copy in export_dir dir.', color='orange')

    files_and_dirs_to_add_map: dict[Path, Path] = _get_enabled_files_and_dirs_private_to_public(
        course_config,
        course_schedule,
        public_course_driver,
        private_course_driver,
    )

    if not dry_run:
        if not course_config.gitlab_service_token:
            raise Exception('Unable to find service_token')  # TODO: set exception correct type

        print_info('Setting up public repo...', color='orange')
        print_info(f'  Copy {course_config.gitlab_url}/{course_config.public_repo} repo in {export_dir}')
        print_info(
            f'  username {course_config.gitlab_service_username} \n'
            f'  name {course_config.gitlab_service_name} \n'
            f'  branch {course_config.default_branch} \n',
            color='grey'
        )
        setup_repo_in_dir(
            export_dir,
            f'{course_config.gitlab_url}/{course_config.public_repo}',
            service_username=course_config.gitlab_service_username,
            service_token=course_config.gitlab_service_token,
            git_user_email=course_config.gitlab_service_email,
            git_user_name=course_config.gitlab_service_name,
            branch=course_config.default_branch,
        )

    # remove all files from export_dir (to delete deleted files)
    print_info('Delete all files from old export_dir (keep .git)...', color='orange')
    deleted_files: set[str] = set()
    for path in export_dir.glob('*'):
        if path.name == '.git':
            continue

        if path.is_file() or path.is_symlink():
            path.unlink()
        elif path.is_dir():
            shutil.rmtree(path)
        else:
            print(f'wtf. {path}')

        deleted_files.add(str(path.as_posix()))

    # copy updated files
    print_info('Copy updated files...', color='orange')
    for filename_private, filename_public in sorted(files_and_dirs_to_add_map.items()):
        relative_private_filename = str(filename_private.relative_to(private_course_driver.root_dir))
        relative_public_filename = str(filename_public.relative_to(public_course_driver.root_dir))
        print_info(f'  {relative_private_filename}', color='grey')
        print_info(f'  \t-> {relative_public_filename}', color='grey')

        if filename_private.is_dir():
            shutil.copytree(filename_private, export_dir / relative_public_filename, dirs_exist_ok=True)
        else:
            (export_dir / relative_public_filename).parent.mkdir(exist_ok=True, parents=True)
            shutil.copy(filename_private, export_dir / relative_public_filename)

    if not dry_run:
        # files for git add
        commit_push_all_repo(
            export_dir,
            branch=course_config.default_branch,
        )



from __future__ import annotations

import shutil
import subprocess
from pathlib import Path

from .print import print_info


DEFAULT_BRANCH = 'main'


def get_tracked_files_list(
        repo_dir: Path,
) -> list[str]:
    r = subprocess.run(
        'git ls-files',
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        check=True,
        cwd=repo_dir,
    )
    return r.stdout.splitlines()


def setup_repo_in_dir(
        repo_dir: Path,
        remote_repo_url: str,
        service_username: str,
        service_token: str,
        git_user_email: str = 'no-reply@gitlab.manytask.org',
        git_user_name: str = 'Manytask Bot',
        branch: str = DEFAULT_BRANCH,
) -> None:
    remote_repo_url = remote_repo_url.replace('https://', '').replace('http://', '').replace('.git', '')
    remote_repo_url = f'http://{service_username}:{service_token}@{remote_repo_url}.git'
    print_info(f'remote_repo_url {remote_repo_url}', color='grey')

    shutil.rmtree(repo_dir)
    repo_dir.mkdir()

    print_info('* git clone...')
    r = subprocess.run(
        f'git clone --depth=1 --branch={branch} {remote_repo_url} ./',
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        cwd=repo_dir,
    )
    print_info(r.stdout, color='grey')
    if r.returncode != 0:
        print_info('empty repo, init it first')
        r = subprocess.run(
            f'git init --initial-branch {branch} && git remote add origin {remote_repo_url} && git fetch',
            encoding='utf-8',
            stdout=subprocess.PIPE,
            shell=True,
            check=True,
            cwd=repo_dir,
        )
        print_info(r.stdout, color='grey')

    print_info('* git remote...')
    r = subprocess.run(
        'git remote -v',
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        cwd=repo_dir,
    )
    print_info(r.stdout, color='grey')

    print_info('* git config set...')
    r = subprocess.run(
        f'git config --local user.email "{git_user_email}" && '
        f'git config --local user.name "{git_user_name}"',
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        cwd=repo_dir,
    )
    print_info(r.stdout, color='grey')


def commit_push_all_repo(
        repo_dir: Path,
        branch: str = DEFAULT_BRANCH,
        message: str = 'Export public files',
) -> None:
    print_info('* git status...')
    r = subprocess.run(
        'git status',
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        check=True,
        cwd=repo_dir,
    )
    print_info(r.stdout, color='grey')

    print_info('* adding files...')
    r = subprocess.run(
        'git add .',
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        cwd=repo_dir,
    )
    print_info(r.stdout, color='grey')

    print_info('* committing...')
    r = subprocess.run(
        f'git commit --all -m "{message}"',
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        cwd=repo_dir,
    )
    print_info(r.stdout, color='grey')

    print_info('* git pushing...')
    r = subprocess.run(
        f'git push -o ci.skip origin {branch}',
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        cwd=repo_dir,
    )
    print_info(r.stdout, color='grey')
    if r.returncode != 0:
        raise Exception('Can not export files in public repo')  # TODO: make right type

    print_info('Done.')



from __future__ import annotations

import re
from pathlib import Path


PRECOMPILED_REGEXPS = {}


def cut_marked_code_from_string(
        content: str,
        clear_mark: str | tuple[str, str],
        clear_mark_replace: str,
        raise_not_found: bool = False,
) -> str:
    """
    Cut the marked source code from the gold solution inplace

    @param content: code to inplace replace code
    @param clear_mark: string or (string, string) to cut from gold solution
    @param clear_mark_replace: string to replace cut code
    @param raise_not_found: raise assertion error if no paired <clear_mark> found
    @return: clear content or original content if no <clear_mark> found
    @raises: assertion error if no paired CLEAR_MARK found

    Cut all content between pairs or CLEAR_MARK strings.
    given clear_mark="YOUR CODE" or clear_mark=("YOUR CODE", "YOUR CODE")
    ```python
    a = 1
    # YOUR CODE
    b = 1
    # YOUR CODE
    ```
    You'll get the following result
    ```python
    a = 1
    # YOUR CODE
    ```
    """
    global PRECOMPILED_REGEXPS
    clear_mark_start, clear_mark_end = clear_mark if isinstance(clear_mark, tuple) else (clear_mark, clear_mark)

    cut_regexp = rf'{clear_mark_start}(.|\s)*?{clear_mark_end}'
    if cut_regexp not in PRECOMPILED_REGEXPS:
        PRECOMPILED_REGEXPS[cut_regexp] = re.compile(cut_regexp)

    template_content, replace_count = re.subn(PRECOMPILED_REGEXPS[cut_regexp], clear_mark_replace, content)

    if raise_not_found and replace_count == 0:
        raise AssertionError(f'Can not find "{clear_mark_start}" to "{clear_mark_end}" pair')

    return template_content


def create_template_from_gold_solution(
        source_filename: Path | str,
        target_filename: Path | str | None = None,
        clear_mark: str | tuple[str, str] = 'TODO: CODE HERE',
        clear_mark_replace: str = 'TODO: CODE HERE',
        raise_not_found: bool = False,
) -> bool:
    """
    Cut the marked source code from the gold solution inplace

    @param source_filename: filename to replace code of
    @param target_filename: filename to write replace code, if None - inplace
    @param clear_mark: string or (string, string) to cut from gold solution
    @param clear_mark_replace: string to replace cut code
    @param raise_not_found: raise assertion error if no paired <clear_mark> found
    @return: true if successfully replaces and false if not found
    @raises: assertion error if no paired clear_mark found
    @see: cut_marked_code_from_string
    """
    source_filename = Path(source_filename)
    assert source_filename.exists() and source_filename.is_file(), f'{source_filename.as_posix()} does not exist'

    target_filename = target_filename or source_filename

    with open(source_filename, 'r') as f:
        content = f.read()

    template_content = cut_marked_code_from_string(
        content,
        clear_mark,
        clear_mark_replace,
        raise_not_found=raise_not_found,
    )

    with open(target_filename, 'w') as f:
        f.write(template_content)

    return content != template_content



from __future__ import annotations

import io
import multiprocessing
import sys
from concurrent.futures import ProcessPoolExecutor, as_completed
from contextlib import redirect_stderr, redirect_stdout

from ..course import CourseDriver, CourseSchedule, Task
from ..exceptions import RunFailedError
from ..testers import Tester
from ..utils.print import print_info, print_task_info


def _check_single_task(
        task: Task,
        tester: Tester,
        private_course_driver: CourseDriver,
        verbose: bool = False,
        catch_output: bool = False,
) -> str | None:
    reference_source_dir = private_course_driver.get_task_solution_dir(task)
    reference_config_dir = private_course_driver.get_task_config_dir(task)
    reference_public_tests_dir = private_course_driver.get_task_public_test_dir(task)
    reference_private_tests_dir = private_course_driver.get_task_private_test_dir(task)
    reference_tests_root_dir = private_course_driver.root_dir
    assert reference_source_dir, 'reference_source_dir have to exists'
    assert reference_config_dir, 'reference_config_dir have to exists'
    assert reference_public_tests_dir or reference_private_tests_dir, \
        'reference_public_tests_dir or reference_private_tests_dir have to exists'

    if catch_output:
        f = io.StringIO()
        with redirect_stderr(f), redirect_stdout(f):
            print_task_info(task.full_name)
            try:
                tester.test_task(
                    reference_source_dir,
                    reference_config_dir,
                    reference_public_tests_dir,
                    reference_private_tests_dir,
                    reference_tests_root_dir,
                    verbose=verbose,
                    normalize_output=True,
                )
            except RunFailedError as e:
                out = f.getvalue()
                raise RunFailedError(e.msg, out + (e.output or '')) from e
            else:
                out = f.getvalue()
                return out
    else:
        print_task_info(task.full_name)
        tester.test_task(
            reference_source_dir,
            reference_config_dir,
            reference_public_tests_dir,
            reference_private_tests_dir,
            reference_tests_root_dir,
            verbose=verbose,
            normalize_output=True,
        )
        return None


def _check_tasks(
        tasks: list[Task],
        tester: Tester,
        private_course_driver: CourseDriver,
        parallelize: bool = False,
        num_processes: int | None = None,
        verbose: bool = True,
) -> bool:
    # Check itself
    if parallelize:
        _num_processes = num_processes or multiprocessing.cpu_count()
        print_info(f'Parallelize task checks with <{_num_processes}> processes...', color='blue')

        success = True
        # with ThreadPoolExecutor(max_workers=num_cores) as e:
        with ProcessPoolExecutor(max_workers=_num_processes) as e:
            check_futures = {
                e.submit(_check_single_task, task, tester, private_course_driver, verbose=verbose, catch_output=True)
                for task in tasks
            }

            for future in as_completed(check_futures):
                try:
                    captured_out = future.result()
                except RunFailedError as e:
                    print_info(e.output)
                    success &= False
                except Exception as e:
                    print_info('Unknown exception:', e, color='red')
                    raise e
                else:
                    print_info(captured_out)
        return success
    else:
        for task in tasks:
            try:
                _check_single_task(task, tester, private_course_driver, verbose=verbose, catch_output=False)
            except RunFailedError:
                return False
            except Exception as e:
                print_info('Unknown exception:', e, color='red')
                raise e

        return True


def pre_release_check_tasks(
        course_schedule: CourseSchedule,
        private_course_driver: CourseDriver,
        tester: Tester,
        tasks: list[Task] | None = None,
        *,
        parallelize: bool = False,
        num_processes: int | None = None,
        contributing: bool = False,
) -> None:
    # select tasks or use `tasks` param
    if tasks:
        print_info('Testing specifying tasks...', color='yellow')
        print_info([i.full_name for i in tasks])
    else:
        if contributing:
            tasks = course_schedule.get_tasks(started=True)
            print_info('Testing started groups...', color='yellow')
            print_info([i.name for i in course_schedule.get_groups(started=True)])
        else:
            tasks = course_schedule.get_tasks(enabled=True)
            print_info('Testing enabled groups...', color='yellow')
            print_info([i.name for i in course_schedule.get_groups(enabled=True)])

    # tests itself
    success = _check_tasks(
        tasks,
        tester,
        private_course_driver,
        parallelize=parallelize,
        num_processes=num_processes,
        verbose=not contributing,
    )

    if not success:
        sys.exit(1)



from __future__ import annotations

import time
from dataclasses import dataclass
from typing import Any

import jinja2.nativetypes

from .configs import PipelineStageConfig
from .exceptions import BadConfig, PluginExecutionFailed
from .plugins import PluginABC
from .utils import print_info


@dataclass
class PipelineStageResult:
    """Result of a single pipeline stage.
    :param name: name of the stage
    :param failed: if True, stage failed
    :param skipped: if True, stage was skipped
    :param percentage: optional percentage of points earned
    :param elapsed_time: optional elapsed time in seconds
    :param output: output of the stage
    """

    name: str
    failed: bool
    skipped: bool
    percentage: float | None = None
    elapsed_time: float | None = None
    output: str = ""

    def __str__(self) -> str:  # pragma: no cover
        return (
            f"PipelineStageResult: failed={int(self.failed)}, "
            f"skipped={int(self.skipped)}, percentage={self.percentage or 1.0:.2f}, name='{self.name}'"
        )


@dataclass
class PipelineResult:
    failed: bool
    stage_results: list[PipelineStageResult]

    def __bool__(self) -> bool:
        return not self.failed

    def __str__(self) -> str:  # pragma: no cover
        return f"PipelineResult: failed={int(self.failed)}\n" + "\n".join(
            [f"  {stage_result}" for stage_result in self.stage_results]
        )


class ParametersResolver:
    def __init__(self) -> None:
        self.template_env = jinja2.nativetypes.NativeEnvironment(
            loader=jinja2.BaseLoader(),
            variable_start_string="${{",
            variable_end_string="}}",
        )

    def resolve(self, template: str | list[str] | Any, context: dict[str, Any]) -> Any:
        """
        Resolve the template with context.
        * If template is a string, resolve it with jinja2
        * If template is a list, resolve each element of the list
        * If template is a dict, resolve each value of the dict
        * Otherwise, return the template as is
        :param template: template string to resolve, following jinja2 syntax.
        :param context: context to use for resolving.
        :return: resolved template of Any type.
        :raises BadConfig: if template is invalid.
        """
        if isinstance(template, str):
            try:
                template_obj = self.template_env.from_string(template.strip())
                return template_obj.render(**context)
            except jinja2.TemplateError as e:
                raise BadConfig(f"Invalid template {template}") from e
        elif isinstance(template, list):
            return [self.resolve(item, context) for item in template]
        elif isinstance(template, dict):
            return {key: self.resolve(value, context) for key, value in template.items()}
        else:
            return template


class PipelineRunner:
    """Class encapsulating the pipeline execution logic."""

    def __init__(
        self,
        pipeline: list[PipelineStageConfig],
        plugins: dict[str, type[PluginABC]],
        *,
        verbose: bool = False,
    ):
        """
        Init pipeline runner with predefined stages/plugins to use, parameters (placeholders) resolved later.
        :param pipeline: list of pipeline stages
        :param plugins: dict of plugins available to use
        :param verbose: if True, print more debug info for teachers
        :raises BadConfig: if plugin does not exist or does not support isolation (no placeholders are checked)
        """
        self.pipeline = pipeline
        self.plugins = plugins

        self.verbose = verbose

        self.parameters_resolver = ParametersResolver()

        self.validate({}, validate_placeholders=False)

    def __len__(self) -> int:
        return len(self.pipeline)

    def validate(
        self,
        context: dict[str, Any],
        validate_placeholders: bool = True,
    ) -> None:
        """
        Validate the pipeline configuration.
        :param context: context to use for resolving placeholders
        :param validate_placeholders: if True, validate placeholders in pipeline stages
        """

        for pipeline_stage in self.pipeline:
            # validate plugin exists
            if pipeline_stage.run not in self.plugins:
                raise BadConfig(f"Unknown plugin {pipeline_stage.run} in pipeline stage {pipeline_stage.name}")
            plugin_class = self.plugins[pipeline_stage.run]

            # validate args of the plugin (first resolve placeholders)
            if validate_placeholders:
                resolved_args = self.parameters_resolver.resolve(pipeline_stage.args, context)
                plugin_class.validate(resolved_args)

            # validate run_if condition
            if validate_placeholders and pipeline_stage.run_if:
                resolved_run_if = self.parameters_resolver.resolve(pipeline_stage.run_if, context)
                if not isinstance(resolved_run_if, bool):
                    raise BadConfig(
                        f"Invalid run_if condition {pipeline_stage.run_if} in pipeline stage {pipeline_stage.name}"
                    )

            # add output to context if set register parameter
            if pipeline_stage.register_output:
                context.setdefault("outputs", {})[pipeline_stage.register_output] = PipelineStageResult(
                    name=pipeline_stage.name,
                    failed=False,
                    skipped=True,
                    percentage=1.0,
                )

    def run(
        self,
        context: dict[str, Any],
        *,
        dry_run: bool = False,
    ) -> PipelineResult:
        pipeline_stage_results = []
        pipeline_passed = True
        skip_the_rest = False
        for pipeline_stage in self.pipeline:
            # resolve placeholders in arguments
            resolved_args = self.parameters_resolver.resolve(pipeline_stage.args, context=context)
            resolved_run_if = (
                self.parameters_resolver.resolve(pipeline_stage.run_if, context=context)
                if pipeline_stage.run_if is not None
                else None
            )

            # if not verbose and skipping (student mode) dont print anything
            if not dry_run and not self.verbose and (skip_the_rest or resolved_run_if is False):
                continue

            print_info(f'--> Running "{pipeline_stage.name}" stage:', color="orange")
            if self.verbose:
                print_info(f"    run_if: {pipeline_stage.run_if}", color="grey")
                print_info(f"    resolved_run_if: {resolved_run_if}", color="grey")
                print_info(f"    fail: {pipeline_stage.fail}", color="grey")
                print_info(f"    run: {pipeline_stage.run}", color="grey")
                print_info(f"    args: {pipeline_stage.args}", color="grey")
                print_info(f"    resolved_args: {resolved_args}", color="grey")

            # skip the rest of stages if failed before
            if skip_the_rest:
                print_info("skipped! (got error above)", color="blue")
                pipeline_stage_results.append(
                    PipelineStageResult(
                        name=pipeline_stage.name,
                        failed=False,
                        skipped=True,
                    )
                )
                continue

            # resolve run condition if any; skip if run_if=False
            if pipeline_stage.run_if is not None:
                if not resolved_run_if:
                    print_info(f"skipped! (run_if={resolved_run_if})", color="blue")
                    pipeline_stage_results.append(
                        PipelineStageResult(
                            name=pipeline_stage.name,
                            failed=False,
                            skipped=True,
                        )
                    )
                    continue

            # select the plugin to run
            plugin_class = self.plugins[pipeline_stage.run]
            plugin = plugin_class()

            # skip if dry run
            if dry_run:
                print_info("[output here]")
                print_info("dry run!", color="blue")
                pipeline_stage_results.append(
                    PipelineStageResult(
                        name=pipeline_stage.name,
                        failed=False,
                        skipped=False,
                        percentage=1.0,
                    )
                )

                # register output if required
                if pipeline_stage.register_output:
                    context.setdefault("outputs", {})[pipeline_stage.register_output] = pipeline_stage_results[-1]

                continue

            # run the plugin with executor
            _start_time = time.perf_counter()
            try:
                result = plugin.run(resolved_args, verbose=self.verbose)
                _end_time = time.perf_counter()
                print_info(result.output or "[empty output]")
                print_info(f"> elapsed time: {_end_time-_start_time:.2f}s", color="grey")
                print_info("ok!", color="green")
                pipeline_stage_results.append(
                    PipelineStageResult(
                        name=pipeline_stage.name,
                        failed=False,
                        skipped=False,
                        output=result.output,
                        percentage=result.percentage,
                        elapsed_time=_end_time - _start_time,
                    )
                )
            except PluginExecutionFailed as e:
                _end_time = time.perf_counter()
                print_info(e.output or "[empty output]")
                print_info(f"> elapsed time: {_end_time-_start_time:.2f}s", color="grey")
                pipeline_stage_results.append(
                    PipelineStageResult(
                        name=pipeline_stage.name,
                        failed=True,
                        skipped=False,
                        output=e.output or "",
                        percentage=e.percentage,
                        elapsed_time=_end_time - _start_time,
                    )
                )
                if pipeline_stage.fail == PipelineStageConfig.FailType.FAST:
                    print_info("error! (now as fail=fast)", color="red")
                    skip_the_rest = True
                    pipeline_passed = False
                elif pipeline_stage.fail == PipelineStageConfig.FailType.AFTER_ALL:
                    print_info("error! (later as fail=after_all)", color="red")
                    pipeline_passed = False
                elif pipeline_stage.fail == PipelineStageConfig.FailType.NEVER:
                    print_info("error! (ignored as fail=never)", color="red")
                    pass
                else:
                    assert False, f"Unknown fail type {pipeline_stage.fail}"  # pragma: no cover

            # register output if required
            if pipeline_stage.register_output:
                context.setdefault("outputs", {})[pipeline_stage.register_output] = pipeline_stage_results[-1]

        return PipelineResult(
            failed=not pipeline_passed,
            stage_results=pipeline_stage_results,
        )



from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest

from checker.configs import CheckerExportConfig, CheckerStructureConfig, ManytaskConfig
from checker.course import Course
from checker.exceptions import BadStructure
from checker.exporter import Exporter

from .conftest import T_GENERATE_FILE_STRUCTURE


def assert_files_in_folder(folder: Path, expected_files: list[str]) -> None:
    # check if all expected files are in folder
    for file in expected_files:
        assert (folder / file).exists(), f"File {file} not found in {folder}"
    # check no other files are in folder
    for file in folder.glob("**/*"):
        if file.is_dir():
            continue
        assert str(file.relative_to(folder)) in expected_files, f"File {file.relative_to(folder)} not expected"


class TestExporterOnSimple:
    @pytest.fixture()
    def simple_deadlines(self) -> ManytaskConfig:
        return ManytaskConfig(
            version=1,
            settings={
                "course_name": "test",
                "gitlab_base_url": "https://google.com",
                "public_repo": "public",
                "students_group": "students",
            },
            ui={"task_url_template": "https://example.com/$GROUP_NAME/$TASK_NAME"},
            deadlines={
                "timezone": "Europe/Berlin",
                "schedule": [
                    {
                        "group": "no_folder_group",
                        "enabled": True,
                        "start": "2021-01-01 00:00:00",
                        "end": "200d",
                        "tasks": [
                            {"task": "task1", "score": 1},
                            {"task": "task2", "score": 1},
                        ],
                    },
                ],
            },
        )

    @pytest.fixture()
    def simple_structure(self) -> CheckerStructureConfig:
        return CheckerStructureConfig(
            ignore_patterns=[".ignore_me"],
            private_patterns=[".*"],
            public_patterns=["public*"],
        )

    @pytest.fixture()
    def simple_export_config(self) -> CheckerExportConfig:
        return CheckerExportConfig(
            destination="https://example.com",
            templates="search_or_create",
        )

    @pytest.fixture()
    def simple_private_folder(self, tmpdir: Path, generate_file_structure: T_GENERATE_FILE_STRUCTURE) -> Path:
        layout = {
            "task1": {
                ".task.yml": "version: 1\nstructure:\n    ignore_patterns: [extra_ignore_me]\n",
                "test.txt": "Some SOLUTION BEGIN\nHello\nSOLUTION END\n",
                "public_file.py": "print('Hello')\n",
                "extra_ignore_me": "",
            },
            "task2": {
                ".task.yml": "",
                "test.txt": "Some",
                "test.txt.template": "Will replace the file",
                ".private.file": "private\n",
            },
            ".ignore_me": "",
            "just_file": "hey",
            ".private.file": "private\n",
            "public_folder": {"file_in_public_folder": "file"},
            "public_file.py": "print('Hello')\n",
        }
        generate_file_structure(layout, Path(tmpdir / "repository"))
        return Path(tmpdir / "repository")

    @pytest.fixture()
    def simple_student_folder(self, tmpdir: Path, generate_file_structure: T_GENERATE_FILE_STRUCTURE) -> Path:
        layout = {
            "task1": {"test.txt": "Some Changes", "public_file.py": "print('Hello LOL this too')\n"},
            "task2": {"test.txt": "Student changed it"},
            "just_file": "hey",
            "public_folder": {"file_in_public_folder": "try to change"},
            "public_file.py": "print('Hello')\n",
        }
        generate_file_structure(layout, Path(tmpdir / "student"))
        return Path(tmpdir / "student")

    @pytest.fixture()
    def simple_export_folder(self, tmpdir: Path) -> Path:
        return Path(tmpdir / "export")

    @pytest.fixture()
    def simple_exporter(
        self,
        tmpdir: Path,
        simple_private_folder: Path,
        simple_student_folder: Path,
        simple_deadlines: ManytaskConfig,
        simple_structure: CheckerStructureConfig,
        simple_export_config: CheckerExportConfig,
    ) -> Exporter:
        course = Course(
            manytask_config=simple_deadlines,
            repository_root=simple_student_folder,
            reference_root=simple_private_folder,
        )
        return Exporter(
            course,
            simple_structure,
            simple_export_config,
        )

    def test_simple_validate_ok(self, tmpdir: Path, simple_exporter: Exporter) -> None:
        simple_exporter.validate()

    def test_simple_validate_mix_templates_in_task(
        self, tmpdir: Path, simple_exporter: Exporter, simple_private_folder: Path, simple_export_folder: Path
    ) -> None:
        # add other time to existing tasks
        (simple_private_folder / "task1" / "new.txt").touch()
        (simple_private_folder / "task1" / "new.txt.template").touch()
        (simple_private_folder / "task2" / "new.txt").touch()
        (simple_private_folder / "task2" / "new.txt").write_text(
            "Some\n SOLUTION BEGIN\nHello\nSOLUTION END\n", encoding="utf-8"
        )

        with pytest.raises(BadStructure) as exc_info:
            simple_exporter.validate()
        assert "use both" in str(exc_info.value)

    @pytest.mark.parametrize(
        "template_type", [CheckerExportConfig.TemplateType.CREATE, CheckerExportConfig.TemplateType.SEARCH]
    )
    def test_simple_validate_no_templates(
        self,
        tmpdir: Path,
        simple_exporter: Exporter,
        simple_private_folder: Path,
        simple_export_folder: Path,
        template_type: str,
    ) -> None:
        simple_exporter.export_config.templates = template_type

        # delete comment template, make wrong .template file
        (simple_private_folder / "task1" / "test.txt").write_text("Some\n")
        (simple_private_folder / "task2" / "not_original_file.txt.template").touch()

        with pytest.raises(BadStructure) as exc_info:
            simple_exporter.validate()
        assert "Have to include at least" in str(exc_info.value)

    @pytest.mark.parametrize(
        "template_type", [CheckerExportConfig.TemplateType.SEARCH, CheckerExportConfig.TemplateType.SEARCH_OR_CREATE]
    )
    def test_simple_validate_no_original_file_for_templates(
        self,
        tmpdir: Path,
        simple_exporter: Exporter,
        simple_private_folder: Path,
        simple_export_folder: Path,
        template_type: str,
    ) -> None:
        simple_exporter.export_config.templates = template_type

        # delete all templates files
        (simple_private_folder / "task1" / "test.txt").unlink()
        (simple_private_folder / "task1" / "test.txt.template").touch()

        with pytest.raises(BadStructure) as exc_info:
            simple_exporter.validate()
        assert "does not have original" in str(exc_info.value)

    @pytest.mark.parametrize(
        "template_type", [CheckerExportConfig.TemplateType.CREATE, CheckerExportConfig.TemplateType.SEARCH]
    )
    def test_simple_validate_wrong_template_type(
        self, tmpdir: Path, simple_exporter: Exporter, template_type: str
    ) -> None:
        simple_exporter.export_config.templates = template_type

        with pytest.raises(BadStructure) as exc_info:
            simple_exporter.validate()
        assert "Templating set to" in str(exc_info.value)

    @pytest.mark.parametrize(
        "file_content",
        [
            "WRONG_START\nHello\nSOLUTION END\n",
            "SOLUTION BEGIN\nHello\nWRONG\n",
            "Hello\n",
            "SOLUTION BEGIN\nHello\nSOLUTION END\nSOLUTION BEGIN\nHello\nWRONG\n",
            "SOLUTION BEGIN\nHello\nSOLUTION END\n\nHello\nSOLUTION END\n",
            "SOLUTION BEGIN\nHello\nSOLUTION END\nSOLUTION BEGIN\nHello\n",
            "SOLUTION BEGIN\nSOLUTION BEGIN\nHello\nSOLUTION END\nSOLUTION END",
        ],
    )
    def test_simple_validate_wrong_template_comment(
        self,
        tmpdir: Path,
        simple_exporter: Exporter,
        simple_private_folder: Path,
        simple_export_folder: Path,
        file_content: str,
    ) -> None:
        simple_exporter.export_config.templates = CheckerExportConfig.TemplateType.CREATE

        # remove .template and write comment-style template
        (simple_private_folder / "task2" / "test.txt.template").unlink()
        (simple_private_folder / "task2" / "test.txt").write_text(file_content, encoding="utf-8")

        with pytest.raises(BadStructure) as exc_info:
            simple_exporter.validate()
        assert "invalid template comments" in str(exc_info.value) or "does not have template comments" in str(
            exc_info.value
        )

    # TODO: ignore_templates tests
    @pytest.mark.parametrize(
        "file_structure, expected_excluded_paths",
        [
            (
                {},
                [],
            ),
            (
                {"some_folder": {"some_file.txt": "123", "empty_file.py": ""}, "other_file": "123"},
                [],
            ),
            (
                {
                    "some_folder": {
                        "some_file.txt": "123",
                        "not_in_the_root_folder.py": "SOLUTION BEGIN\nSOLUTION END",
                    },
                    "other_file": "123",
                },
                [],  # search in current dir only
            ),
            (
                {
                    "some_folder": {"some_file.txt": "123"},
                    "empty_after_template_file.py": "SOLUTION BEGIN\nSOLUTION END",
                },
                ["empty_after_template_file.py"],
            ),
            (
                {
                    "some_folder": {"some_file.txt": "123"},
                    "empty_after_template_file.py": "   \n\n  SOLUTION BEGIN\n\nSOLUTION END\n\t\n",
                },
                ["empty_after_template_file.py"],
            ),
            (
                {
                    "some_folder": {"some_file.txt.template": "123", "some_file.txt": "321", "empty_file.py": ""},
                    "other_file": "123",
                },
                [],  # search in current dir only
            ),
            (
                {"some_folder": {"empty_file.py": ""}, "some_file.txt.template": "123", "some_file.txt": "321"},
                ["some_file.txt"],
            ),
            (
                {
                    "some_folder.template": {"some_file.txt": "123", "empty_file.py": ""},
                    "some_folder": {"some_file": ""},
                    "other_file": "123",
                },
                ["some_folder"],
            ),
            (
                {
                    "some_folder": {"some_file.txt.template": "123", "some_file.txt": "321", "empty_file.py": ""},
                    "other_file": "SOLUTION BEGIN\nSOLUTION END",
                    "other_other_file": "",
                    "other_other_file.template": "",
                },
                ["other_other_file", "other_file"],
            ),
        ],
    )
    def test_search_for_exclude_due_to_templates(
        self,
        tmpdir: Path,
        simple_exporter: Exporter,
        generate_file_structure: T_GENERATE_FILE_STRUCTURE,
        file_structure: dict[str, Any],
        expected_excluded_paths: list[str],
    ) -> None:
        # use other directory, not original Exporter one
        generate_file_structure(file_structure, root=Path(tmpdir / "test_data"))

        excluded_paths = simple_exporter._search_for_exclude_due_to_templates(Path(tmpdir / "test_data"), False)
        assert sorted(excluded_paths) == sorted(expected_excluded_paths)

    @pytest.mark.parametrize(
        "copy_public, copy_private, copy_other, expected_files",
        [
            (True, False, False, ["task1/public_file.py", "public_file.py", "public_folder/file_in_public_folder"]),
            (False, True, False, ["task1/.task.yml", "task2/.task.yml", "task2/.private.file", ".private.file"]),
            (False, False, True, ["task1/test.txt", "task2/test.txt", "just_file"]),
        ],
    )
    @pytest.mark.parametrize(
        "fill_templates",
        [True, False],
    )
    def test_copy_files_with_config(
        self,
        tmpdir: Path,
        simple_exporter: Exporter,
        simple_private_folder: Path,
        simple_export_folder: Path,
        copy_public: bool,
        copy_private: bool,
        copy_other: bool,
        expected_files: list[str],
        fill_templates: bool,
    ) -> None:
        simple_exporter._copy_files_with_config(
            simple_private_folder,
            simple_export_folder,
            simple_exporter.structure_config,
            copy_public,
            copy_private,
            copy_other,
            fill_templates,
        )

        assert_files_in_folder(simple_export_folder, expected_files)

    @pytest.mark.parametrize("is_file", [True, False])
    def test_copy_files_with_config_empty_template(
        self,
        tmpdir: Path,
        simple_exporter: Exporter,
        simple_private_folder: Path,
        simple_export_folder: Path,
        is_file: bool,
    ) -> None:
        # set .template as empty file/folder will delete original file
        if is_file:
            (simple_private_folder / "task2" / "test.txt.template").write_text("", encoding="utf-8")
        else:
            (simple_private_folder / "task2" / "test.txt.template").unlink()
            (simple_private_folder / "task2" / "test.txt.template").mkdir()

        simple_exporter._copy_files_with_config(
            simple_private_folder,
            simple_export_folder,
            simple_exporter.structure_config,
            False,
            False,
            True,
            True,
        )

        assert_files_in_folder(simple_export_folder, ["task1/test.txt", "just_file"])

    @pytest.mark.parametrize("is_source_file", [True, False])
    @pytest.mark.parametrize("is_destination_file", [True, False])
    def test_copy_files_with_config_template_is_file_or_folder(
        self,
        tmpdir: Path,
        simple_exporter: Exporter,
        simple_private_folder: Path,
        simple_export_folder: Path,
        is_source_file: bool,
        is_destination_file: bool,
    ) -> None:
        # delete original template
        Path(simple_private_folder / "task2" / "test.txt").unlink()
        Path(simple_private_folder / "task2" / "test.txt.template").unlink()

        # make template folder or file
        if is_source_file:
            Path(simple_private_folder / "task2" / "test.txt.template").write_text("NEW TEXT", encoding="utf-8")
        else:
            Path(simple_private_folder / "task2" / "test.txt.template").mkdir()
            Path(simple_private_folder / "task2" / "test.txt.template" / "new_file.txt").write_text(
                "NEW TEXT", encoding="utf-8"
            )

        # make original folder or file
        if is_destination_file:
            Path(simple_private_folder / "task2" / "test.txt").write_text("OLD TEXT", encoding="utf-8")
        else:
            Path(simple_private_folder / "task2" / "test.txt").mkdir()
            Path(simple_private_folder / "task2" / "test.txt" / "old_file.txt").write_text("OLD TEXT", encoding="utf-8")

        simple_exporter._copy_files_with_config(
            simple_private_folder,
            simple_export_folder,
            simple_exporter.structure_config,
            False,
            False,
            True,
            True,
        )

        # regardless of original file - just copy there template
        if is_source_file:
            assert (simple_export_folder / "task2" / "test.txt").is_file()
            assert (simple_export_folder / "task2" / "test.txt").read_text(encoding="utf-8") == "NEW TEXT"
        else:
            assert (simple_export_folder / "task2" / "test.txt").is_dir()
            assert (simple_export_folder / "task2" / "test.txt" / "new_file.txt").is_file()
            assert (simple_export_folder / "task2" / "test.txt" / "new_file.txt").read_text(
                encoding="utf-8"
            ) == "NEW TEXT"

    def test_export_public(
        self, tmpdir: Path, simple_exporter: Exporter, simple_private_folder: Path, simple_export_folder: Path
    ) -> None:
        simple_exporter.export_public(simple_export_folder)

        assert_files_in_folder(
            simple_export_folder.resolve(),
            [
                "task1/test.txt",
                "task1/public_file.py",
                "task2/test.txt",
                "just_file",
                "public_folder/file_in_public_folder",
                "public_file.py",
            ],
        )
        # check templates was resolved if needed (all here)
        assert (simple_export_folder / "task1" / "test.txt").read_text(encoding="utf-8") == "Some TODO: Your solution\n"
        assert (simple_export_folder / "task2" / "test.txt").read_text(encoding="utf-8") == "Will replace the file"

    def test_export_for_testing(
        self, tmpdir: Path, simple_exporter: Exporter, simple_private_folder: Path, simple_export_folder: Path
    ) -> None:
        simple_exporter.export_for_testing(simple_export_folder)

        assert_files_in_folder(
            simple_export_folder.resolve(),
            [
                "task1/.task.yml",
                "task1/test.txt",
                "task1/public_file.py",
                "task2/.task.yml",
                "task2/test.txt",
                "task2/.private.file",
                ".private.file",
                "just_file",
                "public_folder/file_in_public_folder",
                "public_file.py",
            ],
        )
        # check templates was resolved if needed (non here)
        assert (simple_export_folder / "task1" / "test.txt").read_text(encoding="utf-8") == "Some Changes"
        assert (simple_export_folder / "task2" / "test.txt").read_text(encoding="utf-8") == "Student changed it"
        assert (simple_export_folder / "just_file").read_text(encoding="utf-8") == "hey"
        # overwritten to public tests
        assert (simple_export_folder / "public_folder" / "file_in_public_folder").read_text(encoding="utf-8") == "file"
        assert (simple_export_folder / "task1" / "public_file.py").read_text(encoding="utf-8") == "print('Hello')\n"

    def test_export_for_contribution(
        self, tmpdir: Path, simple_exporter: Exporter, simple_private_folder: Path, simple_export_folder: Path
    ) -> None:
        simple_exporter.export_for_contribution(simple_export_folder)

        assert_files_in_folder(
            simple_export_folder.resolve(),
            [
                "task1/.task.yml",
                "task1/test.txt",
                "task1/public_file.py",
                "task2/.task.yml",
                "task2/test.txt",
                "task2/.private.file",
                ".private.file",
                "just_file",
                "public_folder/file_in_public_folder",
                "public_file.py",
            ],
        )
        # check templates was resolved if needed (non here)
        assert (simple_export_folder / "task1" / "test.txt").read_text(
            encoding="utf-8"
        ) == "Some SOLUTION BEGIN\nHello\nSOLUTION END\n"
        assert (simple_export_folder / "task2" / "test.txt").read_text(encoding="utf-8") == "Some"
        assert (simple_export_folder / "public_folder" / "file_in_public_folder").read_text(
            encoding="utf-8"
        ) == "try to change"
        assert (simple_export_folder / "task1" / "public_file.py").read_text(
            encoding="utf-8"
        ) == "print('Hello LOL this too')\n"


class _TestExporter:
    SAMPLE_TEST_DEADLINES_CONFIG = ManytaskConfig(
        version=1,
        settings={
            "course_name": "test",
            "gitlab_base_url": "https://google.com",
            "public_repo": "public",
            "students_group": "students",
        },
        ui={"task_url_template": "https://example.com/$GROUP_NAME/$TASK_NAME"},
        deadlines={
            "timezone": "Europe/Berlin",
            "schedule": [
                {
                    "group": "group",
                    "enabled": True,
                    "start": "2021-01-01 00:00:00",
                    "end": "200d",
                    "tasks": [
                        {"task": "task1", "score": 1},
                        {"task": "task2", "score": 1},
                        {"task": "task3", "enabled": False, "score": 1},
                    ],
                },
                {
                    "group": "disabled_group",
                    "enabled": False,
                    "start": "2021-01-01 00:00:00",
                    "end": "200d",
                    "tasks": [
                        {"task": "task_disabled_1", "score": 1},
                        {"task": "task_disabled_2", "enabled": True, "score": 1},
                    ],
                },
                {
                    "group": "no_folder_group",
                    "enabled": True,
                    "start": "2021-01-01 00:00:00",
                    "end": "200d",
                    "tasks": [{"task": "root_task_1", "score": 1}],
                },
            ],
        },
    )
    SAMPLE_TEST_STRUCTURE_CONFIG = CheckerStructureConfig(
        ignore_patterns=[".ignore_folder"],
        public_patterns=[".private_exception", ".group.yml"],  # note: .task.yml ignored by default
        private_patterns=[".*", "private.*"],
    )
    SAMPLE_EXPORT_CONFIG = CheckerExportConfig(
        destination="https://example.com",
        templates="search_or_create",
    )
    SAMPLE_TEST_FILES = {
        ".ignore_folder": {
            "folder": {
                "test.txt": "Hello2\n",
                "test.py": "print('Hello2')\n",
            },
        },
        ".private_folder": {
            "test.txt": "Hello3\n",
            "folder": {
                ".test.py": "print('Hello3')\n",
                "test.txt": "Hello4\n",
            },
        },
        "folder": {
            "test.txt": "Hello2\n",
            ".test.py": "print('Hello2')\n",
            "folder": {
                "test.txt": "Hello2\n",
            },
        },
        "other_folder": {
            "test.txt": "Hello5\n",
        },
        "group": {
            "task1": {
                ".task.yml": "version: 1\nstructure:\n    private_patterns: []\n",
                "test.txt": "SOLUTION BEGIN\nHello\nSOLUTION END\n",
                ".test.py": "print('Hello')\n",  # not private anymore, override
            },
            "task2": {
                "junk_group_folder": {"some_junk_file.txt": "Junk\n"},
                "junk_group_folder.template": "",  # will delete junk_group_folder
                ".task.yml": "version: 1",
                "private.txt": "Private\n",
                "private.py": "print('Private')\n",
                "valid.txt": "Valid\n",
                "valid.txt.template": "Valid Template\n",
            },
            "junk_file.py": "123",
            ".group.yml": "version: 1\nstructure:\n    ignore_patterns: [junk_group_folder, junk_file.py]\n",
        },
        "root_task_1": {
            ".task.yml": "version: 1\nstructure:\n    public_patterns: []\n",
            ".private_exception": "Some line\n",  # not public anymore, override
            "test.txt": "Hello\n",
            "test.txt.template": "Hello Template\n",
        },
        "test.py": "print('Hello')\n",
        "test.txt": "Hello\n",
        ".some_file": "Some line\n",
        ".private_exception": "Some line\n",
        "private.txt": "Private\n",
        "private.py": "print('Private')\n",
    }

    def test_validate_ok_simple(
        self,
        tmpdir: Path,
        generate_file_structure: T_GENERATE_FILE_STRUCTURE,
        simple_private_folder: Path,
        simple_export_folder: Path,
    ) -> None:
        structure_config = CheckerStructureConfig(
            ignore_patterns=[".gitignore"],
            private_patterns=[".*"],
            public_patterns=["*"],
        )
        generate_file_structure(
            {
                "task1": {".task.yml": "version: 1\n", "test.txt": "SOLUTION BEGIN\nHello\nSOLUTION END\n"},
                "test.py": "print('Hello')\n",
            },
            root=simple_private_folder,
        )
        course = Course(
            manytask_config=ManytaskConfig(
                version=1,
                settings={
                    "course_name": "test",
                    "gitlab_base_url": "https://google.com",
                    "public_repo": "public",
                    "students_group": "students",
                },
                ui={"task_url_template": "https://example.com/$GROUP_NAME/$TASK_NAME"},
                deadlines={
                    "timezone": "Europe/Berlin",
                    "schedule": [
                        {
                            "group": "no_folder_group",
                            "enabled": True,
                            "start": "2021-01-01 00:00:00",
                            "end": "200d",
                            "tasks": [
                                {"task": "task1", "score": 1},
                            ],
                        },
                    ],
                },
            ),
            repository_root=simple_private_folder,
        )
        exporter = Exporter(
            course,
            structure_config,
            self.SAMPLE_EXPORT_CONFIG,
            simple_private_folder,
        )
        exporter.validate()

    def test_search_for_exclude_due_templates(
        self,
        tmpdir: Path,
        generate_file_structure: T_GENERATE_FILE_STRUCTURE,
        simple_private_folder: Path,
        simple_export_folder: Path,
        file_structure: dict[str, Any],
        expected_excluded_paths: list[str],
    ) -> None:
        generate_file_structure(self.SAMPLE_TEST_FILES, root=simple_private_folder)
        course = Course(
            manytask_config=self.SAMPLE_TEST_DEADLINES_CONFIG,
            repository_root=simple_private_folder,
        )
        exporter = Exporter(
            course,
            self.SAMPLE_TEST_STRUCTURE_CONFIG,
            self.SAMPLE_EXPORT_CONFIG,
            simple_private_folder,
        )

        generate_file_structure(file_structure, root=Path(tmpdir / "test_data"))
        excluded_paths = exporter._search_for_exclude_due_to_templates(Path(tmpdir / "test_data"), False)
        assert sorted(path_name for path_name in excluded_paths) == sorted(expected_excluded_paths)

    @pytest.mark.parametrize(
        "file_content",
        [
            "TEM!PLATE START\nHello\nSOLUTION END\n",
            "SOLUTION BEGIN\nHello\nTEMPL!ATE END\n",
            "Hello\n",
            "SOLUTION BEGIN\nHello\nSOLUTION END\nSOLUTION BEGIN\nHello\nTEMP!LATE END\n",
            "SOLUTION BEGIN\nHello\nSOLUTION END\n\nHello\nSOLUTION END\n",
            "SOLUTION BEGIN\nHello\nSOLUTION END\nSOLUTION BEGIN\nHello\n",
            "SOLUTION BEGIN\nSOLUTION BEGIN\nHello\nSOLUTION END\nSOLUTION END",
        ],
    )
    def test_validate_template_wrong_template_comments(
        self,
        tmpdir: Path,
        generate_file_structure: T_GENERATE_FILE_STRUCTURE,
        simple_private_folder: Path,
        simple_export_folder: Path,
        file_content: str,
    ) -> None:
        structure_config = CheckerStructureConfig(
            ignore_patterns=[".gitignore"],
            private_patterns=[".*"],
            public_patterns=["*"],
        )
        generate_file_structure(
            {
                "task1": {".task.yml": "version: 1\n", "test.txt": file_content},
                "test.py": "print('Hello')\n",
            },
            root=simple_private_folder,
        )
        course = Course(
            manytask_config=ManytaskConfig(
                version=1,
                settings={
                    "course_name": "test",
                    "gitlab_base_url": "https://google.com",
                    "public_repo": "public",
                    "students_group": "students",
                },
                ui={"task_url_template": "https://example.com/$GROUP_NAME/$TASK_NAME"},
                deadlines={
                    "timezone": "Europe/Berlin",
                    "schedule": [
                        {
                            "group": "no_folder_group",
                            "enabled": True,
                            "start": "2021-01-01 00:00:00",
                            "end": "200d",
                            "tasks": [
                                {"task": "task1", "score": 1},
                            ],
                        },
                    ],
                },
            ),
            repository_root=simple_private_folder,
        )
        exporter = Exporter(
            course,
            structure_config,
            CheckerExportConfig(destination="https://example.com", templates="create"),
            simple_private_folder,
        )
        with pytest.raises(BadStructure):
            exporter.validate()

    def test_validate_template_search_ok(
        self,
        tmpdir: Path,
        generate_file_structure: T_GENERATE_FILE_STRUCTURE,
        simple_private_folder: Path,
        simple_export_folder: Path,
    ) -> None:
        structure_config = CheckerStructureConfig(
            ignore_patterns=[".gitignore"],
            private_patterns=[".*"],
            public_patterns=["*"],
        )
        generate_file_structure(
            {
                "folder": {"test.txt.template": "Hello\n"},
                "test.py": "print('Hello')\n",
            },
            root=simple_private_folder,
        )
        course = Course(
            manytask_config=self.SAMPLE_TEST_DEADLINES_CONFIG,
            repository_root=simple_private_folder,
        )
        exporter = Exporter(
            course,
            structure_config,
            self.SAMPLE_EXPORT_CONFIG,
            simple_private_folder,
        )

        exporter.validate()

    def test_export_public(
        self,
        tmpdir: Path,
        generate_file_structure: T_GENERATE_FILE_STRUCTURE,
        simple_private_folder: Path,
        simple_export_folder: Path,
    ) -> None:
        generate_file_structure(self.SAMPLE_TEST_FILES, root=simple_private_folder)
        course = Course(
            manytask_config=self.SAMPLE_TEST_DEADLINES_CONFIG,
            repository_root=simple_private_folder,
        )
        exporter = Exporter(
            course,
            self.SAMPLE_TEST_STRUCTURE_CONFIG,
            self.SAMPLE_EXPORT_CONFIG,
            simple_private_folder,
        )

        exporter.export_public(simple_export_folder)

        assert_files_in_folder(
            simple_export_folder.resolve(),
            [
                "folder/test.txt",
                "folder/folder/test.txt",
                "other_folder/test.txt",
                "test.py",
                "test.txt",
                ".private_exception",
                "group/task1/.task.yml",
                "group/task1/test.txt",
                "group/task1/.test.py",
                "group/task2/valid.txt",
                "group/.group.yml",
                "root_task_1/test.txt",
            ],
        )

    def test_export_for_testing(
        self,
        tmpdir: Path,
        generate_file_structure: T_GENERATE_FILE_STRUCTURE,
        simple_private_folder: Path,
        simple_export_folder: Path,
    ) -> None:
        generate_file_structure(self.SAMPLE_TEST_FILES, root=simple_private_folder)
        course = Course(
            manytask_config=self.SAMPLE_TEST_DEADLINES_CONFIG,
            repository_root=simple_private_folder,
        )
        exporter = Exporter(
            course,
            self.SAMPLE_TEST_STRUCTURE_CONFIG,
            self.SAMPLE_EXPORT_CONFIG,
            simple_private_folder,
        )

        exporter.export_for_testing(simple_export_folder)

        assert_files_in_folder(
            simple_export_folder.resolve(),
            [
                ".private_folder/test.txt",
                ".private_folder/folder/.test.py",
                ".private_folder/folder/test.txt",
                "folder/test.txt",
                "folder/.test.py",
                "folder/folder/test.txt",
                "other_folder/test.txt",
                "test.py",
                "test.txt",
                ".some_file",
                ".private_exception",
                "private.txt",
                "private.py",
                "group/task1/.task.yml",
                "group/task1/test.txt",
                "group/task1/.test.py",
                "group/task2/.task.yml",
                "group/task2/private.txt",
                "group/task2/private.py",
                "group/task2/valid.txt",
                "group/.group.yml",
                "root_task_1/.task.yml",
                "root_task_1/test.txt",
                "root_task_1/.private_exception",
            ],
        )

    def test_export_for_contribution(
        self,
        tmpdir: Path,
        generate_file_structure: T_GENERATE_FILE_STRUCTURE,
        simple_private_folder: Path,
        simple_export_folder: Path,
    ) -> None:
        generate_file_structure(self.SAMPLE_TEST_FILES, root=simple_private_folder)
        course = Course(
            manytask_config=self.SAMPLE_TEST_DEADLINES_CONFIG,
            repository_root=simple_private_folder,
        )
        exporter = Exporter(
            course,
            self.SAMPLE_TEST_STRUCTURE_CONFIG,
            self.SAMPLE_EXPORT_CONFIG,
            simple_private_folder,
        )

        exporter.export_for_contribution(simple_export_folder)

        assert_files_in_folder(
            simple_export_folder.resolve(),
            [
                ".private_folder/test.txt",
                ".private_folder/folder/.test.py",
                ".private_folder/folder/test.txt",
                "folder/test.txt",
                "folder/.test.py",
                "folder/folder/test.txt",
                "other_folder/test.txt",
                "test.py",
                "test.txt",
                ".some_file",
                ".private_exception",
                "private.txt",
                "private.py",
                "group/task1/.task.yml",
                "group/task1/test.txt",
                "group/task1/.test.py",
                "group/task2/.task.yml",
                "group/task2/private.txt",
                "group/task2/private.py",
                "group/task2/valid.txt",
                "group/.group.yml",
                "root_task_1/.task.yml",
                "root_task_1/test.txt",
                "root_task_1/.private_exception",
            ],
        )



from __future__ import annotations

import os
import subprocess
import sys
import tempfile
from datetime import datetime
from pathlib import Path

from ..course import CourseConfig, CourseDriver, CourseSchedule, Group, Task
from ..exceptions import RunFailedError
from ..testers import Tester
from ..utils import get_folders_diff_except_public, get_tracked_files_list
from ..utils.manytask import PushFailedError, push_report
from ..utils.print import print_info, print_task_info


class GitException(Exception):
    pass


def _get_git_changes(
        solution_root: str,
        public_repo_url: str,
        author_name: str | None = None,
        current_commit_sha: str | None = None,
        prev_commit_sha: str | None = None,
        git_changes_type: str = 'log_between_no_upstream',
) -> list[str]:
    """
    :param solution_root: Full path to solutions folder
    :param public_repo_url: Full url to public repo
    :param git_changes_type: one of
        'diff_last', 'diff_between', 'log_between_no_merges', 'log_between_by_author', 'log_between_no_upstream'
    """
    if author_name is None and git_changes_type == 'log_between_by_author':
        git_changes_type = 'log_between_no_merges'

    if prev_commit_sha and set(prev_commit_sha) == {'0'}:  # first commit or merge request
        prev_commit_sha = None

    if 'between' in git_changes_type and (current_commit_sha is None or prev_commit_sha is None):
        print_info('CI_COMMIT_SHA or CI_COMMIT_BEFORE_SHA is wrong pipeline_diff can not be used. '
                   'Using std `git show`')
        print_info(f'CI_COMMIT_SHA: {current_commit_sha}, CI_COMMIT_BEFORE_SHA: {prev_commit_sha}!')
        git_changes_type = 'diff_last'

    changes = []
    if git_changes_type.startswith('diff'):
        if git_changes_type == 'diff_between':
            print_info(f'Looking diff between {prev_commit_sha} and {current_commit_sha}...')
            prev_commit_sha = '' if prev_commit_sha is None else prev_commit_sha
            git_status = subprocess.run(
                f'cd {solution_root} && git diff {prev_commit_sha} {current_commit_sha} --stat --oneline',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                shell=True
            ).stdout
            print_info(git_status)
            changes = git_status.split('\n')[:-2]
        elif git_changes_type == 'diff_last':
            print_info('Looking last commit diff...')
            git_status = subprocess.run(
                f'cd {solution_root} && git show --stat --oneline',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                shell=True
            ).stdout
            print_info(git_status)
            changes = git_status.split('\n')[1:-2]
        else:
            raise GitException(f'Unknown git_changes_type={git_changes_type}')

        changes = [f.rsplit('|', maxsplit=1)[0].strip() for f in changes]

    elif git_changes_type.startswith('log'):
        if git_changes_type == 'log_between_no_merges':
            print_info(f'Looking log between {prev_commit_sha} and {current_commit_sha} without merges...')
            prev_commit_sha = '' if prev_commit_sha is None else prev_commit_sha
            git_status = subprocess.run(
                f'cd {solution_root} && '
                f'git log --pretty="%H" --no-merges {prev_commit_sha or ""}..{current_commit_sha} | '
                f'while read commit_hash; do git show --oneline --name-only $commit_hash'
                f'| tail -n+2; done | sort | uniq',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                shell=True
            ).stdout
            print_info(git_status)
            changes = git_status.split('\n')
        elif git_changes_type == 'log_between_by_author':
            assert isinstance(author_name, str)
            print_info(
                f'Looking log between {prev_commit_sha} and {current_commit_sha} '
                f'by author="{author_name.split(" ")[0]}"...',
            )
            prev_commit_sha = '' if prev_commit_sha is None else prev_commit_sha
            git_status = subprocess.run(
                f'cd {solution_root} && '
                f'git log --pretty="%H" --author="{author_name.split(" ")[0]}"'
                f'{prev_commit_sha or ""}..{current_commit_sha} | '
                f'while read commit_hash; do git show --oneline --name-only $commit_hash '
                f'| tail -n+2; done | sort | uniq',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                shell=True,
            ).stdout
            print_info(git_status)
            changes = git_status.split('\n')
        elif git_changes_type == 'log_between_no_upstream':
            print_info(f'Looking log_between_no_upstream between {prev_commit_sha} and {current_commit_sha} '
                       f'which not in `{public_repo_url}`...')

            result = subprocess.run(
                f'cd {solution_root} && '
                f'git fetch --unshallow &&'
                f'(git remote rm upstream | true) &&'
                f'git remote add upstream {public_repo_url}.git &&'
                f'git fetch upstream',
                encoding='utf-8',
                capture_output=True,
                shell=True
            )
            print_info(result.stderr, color='grey')
            print_info(result.stdout, color='grey')

            print_info('---')

            git_status = subprocess.run(
                f'cd {solution_root} && '
                f'git log --pretty="%H" {prev_commit_sha or ""}..{current_commit_sha} '
                f'--no-merges --not --remotes=upstream | '
                f'while read commit_hash; do git show --oneline --name-only $commit_hash '
                f'| tail -n+2; done | sort | uniq',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                shell=True
            ).stdout
            print_info('Detected changes in the following files:')
            print_info(git_status, color='grey')
            changes = git_status.split('\n')
        else:
            raise GitException(f'Unknown git_changes_type={git_changes_type}')

        changes = [f for f in changes if len(f) > 0]

    return changes


def grade_single_task(
        task: Task,
        tester: Tester,
        course_config: CourseConfig,
        public_course_driver: CourseDriver,
        private_course_driver: CourseDriver,
        user_id: int,
        send_time: datetime,
        inspect: bool = False
) -> bool:
    print_task_info(task.full_name)
    source_dir = public_course_driver.get_task_solution_dir(task)
    reference_config_dir = private_course_driver.get_task_config_dir(task)
    reference_public_tests_dir = private_course_driver.get_task_public_test_dir(task)
    reference_private_tests_dir = private_course_driver.get_task_private_test_dir(task)
    reference_tests_root_dir = private_course_driver.root_dir
    assert source_dir, 'source_dir have to exists'
    assert reference_config_dir, 'reference_config_dir have to exists'
    assert reference_public_tests_dir or reference_private_tests_dir, \
        'reference_public_tests_dir or reference_private_tests_dir have to exists'

    try:
        score_percentage = tester.test_task(
            source_dir,
            reference_config_dir,
            reference_public_tests_dir,
            reference_private_tests_dir,
            reference_tests_root_dir,
            verbose=inspect,
            normalize_output=inspect,
        )
        score = round(score_percentage * task.max_score)
        if score_percentage == 1.:
            print_info(f'\nSolution score is: {score}', color='green')
        else:
            print_info(f'\nSolution score percentage is: {score_percentage}', color='green')
            print_info(f'\nSolution score is: [{task.max_score}*{score_percentage}]={score}', color='green')
        if task.review:
            print_info('\nThis task is "review-able", so, open MR and wait till review.', color='blue')
        elif not inspect:
            use_demand_multiplier = not task.marked
            try:
                if not course_config.manytask_token:
                    raise PushFailedError('Unable to find manytask token')
                files = {
                    path.name: (str(path.relative_to(source_dir)), open(path, 'rb'))
                    for path in source_dir.glob('**/*')
                    if path.is_file()
                }
                username, set_score, result_commit_time, result_submit_time, demand_multiplier = push_report(
                    course_config.manytask_url,
                    course_config.manytask_token,
                    task.name,
                    user_id,
                    score,
                    files=files,
                    send_time=send_time,
                    use_demand_multiplier=use_demand_multiplier,
                )
                print_info(
                    f'Final score for @{username} (according to deadlines and demand): {set_score}',
                    color='blue'
                )
                if demand_multiplier and demand_multiplier != 1:
                    print_info(
                        f'Due to low demand, the task score is multiplied at {demand_multiplier:.4f}',
                        color='grey'
                    )
                if result_commit_time:
                    print_info(f'Commit at {result_commit_time} (are validated to Submit Time)', color='grey')
                if result_submit_time:
                    print_info(f'Submit at {result_submit_time} (deadline is calculated relative to it)', color='grey')
            except PushFailedError:
                raise
        return True
    except RunFailedError:
        # print_info(e)
        return False


def grade_tasks(
        tasks: list[Task],
        tester: Tester,
        course_config: CourseConfig,
        public_course_driver: CourseDriver,
        private_course_driver: CourseDriver,
        user_id: int,
        send_time: datetime,
        inspect: bool = False
) -> bool:
    success = True
    for task in tasks:
        success &= grade_single_task(
            task,
            tester,
            course_config,
            public_course_driver,
            private_course_driver,
            user_id,
            send_time,
            inspect=inspect
        )
    return success


def _get_changes_using_real_folders(
        course_config: CourseConfig,
        current_folder: str,
        old_hash: str,
        current_repo_gitlab_path: str,
        gitlab_token: str,
) -> list[str]:
    gitlab_url_with_token = course_config.gitlab_url.replace('://', f'://gitlab-ci-token:{gitlab_token}@')

    with tempfile.TemporaryDirectory() as public_dir:
        with tempfile.TemporaryDirectory() as old_dir:
            # download public repo, minimal
            print_info(f'Cloning {course_config.public_repo} of {course_config.default_branch}...', color='white')
            # print_info('git clone:', color='grey')
            subprocess.run(
                f'git clone --depth=1 --branch={course_config.default_branch} '
                f'{course_config.gitlab_url}/{course_config.public_repo}.git {public_dir}',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True,
            )
            # print_info(r.stdout, color='grey')
            # print_info(f'ls -lah {public_dir}', color='grey')
            subprocess.run(
                f'ls -lah {public_dir}',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True,
            )
            # print_info(r.stdout, color='grey')

            # download old repo by hash, minimal
            print_info(f'Cloning {current_repo_gitlab_path} to get {old_hash}...', color='white')
            # print_info('git clone:', color='grey')
            subprocess.run(
                f'git clone --depth=1 --branch={course_config.default_branch} '
                f'{gitlab_url_with_token}/{current_repo_gitlab_path}.git {old_dir}',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True,
            )
            # print_info(r.stdout, color='grey')
            # print_info(f'git fetch origin {old_hash} && git checkout FETCH_HEAD:', color='grey')
            subprocess.run(
                f'git fetch origin {old_hash} && git checkout FETCH_HEAD',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True,
                cwd=old_dir,
            )
            # print_info(r.stdout, color='grey')
            # print_info(f'ls -lah {old_dir}', color='grey')
            subprocess.run(
                f'ls -lah {old_dir}',
                encoding='utf-8',
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True,
            )
            # print_info(r.stdout, color='grey')

            # get diff
            print_info('Detected changes (filtering by public repo and git tracked files)', color='white')
            print_info('and filtering by git tracked files', color='white')
            changes = get_folders_diff_except_public(
                Path(public_dir),
                Path(old_dir),
                Path(current_folder),
                exclude_patterns=['.git'],
            )
            # filter by tracked by git
            git_tracked_files = get_tracked_files_list(Path(current_folder))
            changes = [f for f in changes if f in git_tracked_files]

            print_info('\nchanged_files:', color='white')
            for change in changes:
                print_info(f'  ->> {change}', color='white')

            return changes


def grade_on_ci(
        course_config: CourseConfig,
        course_schedule: CourseSchedule,
        public_course_driver: CourseDriver,
        private_course_driver: CourseDriver,
        tester: Tester,
        *,
        test_full_groups: bool = False,
        task: list[str] | None = None,
        group: list[str] | None = None,
) -> None:
    solution_root = os.environ['CI_PROJECT_DIR']

    current_time = datetime.now()
    commit_time = datetime.fromisoformat(os.environ['CI_COMMIT_TIMESTAMP'])
    # TODO: check datetime format
    pipeline_created_time: datetime | None = (
        datetime.strptime(os.environ['CI_PIPELINE_CREATED_AT'], '%Y-%m-%dT%H:%M:%SZ')
        if 'CI_PIPELINE_CREATED_AT' in os.environ else None
    )
    job_start_time: datetime | None = (
        datetime.strptime(os.environ['CI_JOB_STARTED_AT'], '%Y-%m-%dT%H:%M:%SZ')
        if 'CI_JOB_STARTED_AT' in os.environ else None
    )
    send_time = pipeline_created_time or current_time

    print_info(f'current_time {current_time}', color='grey')
    print_info(f'-> commit_time {commit_time}', color='grey')
    print_info(f'-> pipeline_created_time {pipeline_created_time}', color='grey')
    print_info(f'-> job_start_time {job_start_time}', color='grey')
    print_info(f'= using send_time {send_time}', color='grey')

    tasks: list[Task] = []

    if not task and not group:
        author_name = os.environ.get('CI_COMMIT_AUTHOR', None)
        current_commit_sha = os.environ.get('CI_COMMIT_SHA', None)
        prev_commit_sha = os.environ.get('CI_COMMIT_BEFORE_SHA', None)
        print_info(f'CI_COMMIT_AUTHOR {author_name}', color='grey')
        print_info(f'CI_COMMIT_SHA {current_commit_sha}', color='grey')
        print_info(f'CI_COMMIT_BEFORE_SHA {prev_commit_sha}', color='grey')


        gitlab_job_token = os.environ.get('CI_JOB_TOKEN') or ''

        print_info('Loading changes...', color='orange')
        # Get changes using real files difference
        try:
            current_repo_gitlab_path = os.environ['CI_PROJECT_PATH']
            changes = _get_changes_using_real_folders(
                course_config,
                current_folder=solution_root,
                old_hash=prev_commit_sha or course_config.default_branch,
                current_repo_gitlab_path=current_repo_gitlab_path,
                gitlab_token=gitlab_job_token,
            )
        except Exception as e:
            print_info('Ooops... Loading changes failed', color='red')
            print_info(e)

            print_info('Trying with git diff instead\n')
            # Get changed files via git
            try:
                changes = _get_git_changes(
                    solution_root,
                    course_config.gitlab_url + '/' + course_config.public_repo,
                    author_name=author_name,
                    current_commit_sha=current_commit_sha,
                    prev_commit_sha=prev_commit_sha,
                )
            except GitException as e:
                print_info('Ooops... Loading changes failed', color='red')
                print_info(e)
                sys.exit(1)

        # Process Changed files to Changed tasks
        groups: list[Group] = []
        for changed_file in changes:
            changed_task_dir = public_course_driver.get_task_dir_name(changed_file)
            if changed_task_dir is None or changed_task_dir not in course_schedule.tasks:
                continue

            # if changed_group_dir == '...':  # if task name is too long it's hidden
            #     changed_group_dir = course_schedule.tasks[changed_task_dir].group.name

            task = course_schedule.tasks[changed_task_dir]
            # group = course.groups[changed_group_dir]
            group = task.group

            # filter tasks and groups
            if group.is_started:
                if task not in tasks:
                    tasks.append(task)
                if group not in groups:
                    groups.append(group)

        # adding all tasks from group to testing
        if test_full_groups:
            print_info('Testing all tasks in changed groups...', color='orange')
            print_info(f'Changed groups: {[i.name for i in groups]}\n')

            tasks = []
            for group in groups:
                tasks.extend(group.tasks)
        else:
            print_info('Testing only changed tasks...', color='orange')
            print_info(f'Changed tasks: {[i.full_name for i in tasks]}\n')
    else:
        # validate tasks and groups
        fs_tasks: dict[str, Task] = dict()
        if task:
            for fs_task in course_schedule.get_tasks(enabled=True):
                if fs_task.name in task:
                    fs_tasks[fs_task.name] = fs_task
        if group:
            for filesystem_group in course_schedule.get_groups(enabled=True):
                if filesystem_group.name in group:
                    for fs_task in filesystem_group.tasks:
                        fs_tasks[fs_task.name] = fs_task
        if fs_tasks:
            print_info(f"Checking tasks: {', '.join(fs_tasks.keys())}")

        if fs_tasks:
            tasks = [task for task in fs_tasks.values()]
        else:
            tasks = None

    # Grade itself
    user_id = int(os.environ['GITLAB_USER_ID'])
    if tasks:
        success = grade_tasks(
            tasks,
            tester,
            course_config,
            public_course_driver,
            private_course_driver,
            user_id=user_id,
            send_time=send_time,
        )
    else:
        print_info('No changed tasks found :(', color='blue')
        print_info('Hint: commit some changes in tasks you are interested in')
        success = False

    if not success:
        sys.exit(1)



from __future__ import annotations

import pytest

from checker.utils import print_ascii_tag, print_header_info, print_info, print_separator


class TestPrint:
    # TODO: test print colors etc

    def test_print_ascii_tag(self, capsys: pytest.CaptureFixture):
        print_ascii_tag()

        captured = capsys.readouterr()
        assert " " in captured.err
        assert "|" in captured.err

    def test_print_info(self, capsys: pytest.CaptureFixture):
        print_info("123")

        captured = capsys.readouterr()
        assert captured.err == "123\n"

    def test_print_separator(self, capsys: pytest.CaptureFixture):
        print_separator("*", string_length=10)

        captured = capsys.readouterr()
        assert "**********" in captured.err

    def test_print_header_info(self, capsys: pytest.CaptureFixture):
        print_header_info("123", string_length=10)

        captured = capsys.readouterr()
        assert "123" in captured.err
        assert "++++++++++" in captured.err



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

from checker.exceptions import (
    BuildFailedError,
    ExecutionFailedError,
    StylecheckFailedError,
    TestsFailedError,
    TimeoutExpiredError,
)

IGNORE_FILE_PATTERNS = ['go.mod', 'go.sum', '.tester.json']

from checker.utils.files import check_files_contains_regexp, copy_files
from checker.utils.print import print_info
from checker.testers.tester import Tester

class GoTester(Tester):

    @dataclass
    class TaskTestConfig(Tester.TaskTestConfig):
        module_name: str = ''
        
        allow_change: list[str] = field(default_factory=list)
        forbidden_regexp: list[str] = field(default_factory=list)
        copy_to_build: list[str] = field(default_factory=list)

        is_crash_me: bool = False

        tests: list[str] = field(default_factory=list)
        input_file: str = ''
        args: list[str] = field(default_factory=list)
        timeout: float = 60.
        capture_output: bool = True

        def __post_init__(
                self,
        ) -> None:
            # assert self.tests
            # assert self.allow_change
            pass

    def _gen_build(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            source_dir: Path,
            public_tests_dir: Path | None,
            private_tests_dir: Path | None,
            tests_root_dir: Path,
            sandbox: bool = True,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> None:
        verbose = true
        check_files_contains_regexp(
            source_dir,
            regexps=test_config.forbidden_regexp,
            patterns=test_config.allow_change,
            raise_on_found=True,
        )
        task_dir = public_tests_dir
        self._executor(
            copy_files,
            source=source_dir,
            target=task_dir,
            patterns=test_config.allow_change,
            verbose=verbose,
        )
        self._executor(
            copy_files,
            source=task_dir,
            target=build_dir,
            patterns=test_config.copy_to_build,
            ignore_patterns=IGNORE_FILE_PATTERNS,
            verbose=verbose,
        )

        try:
            print_info('Init GO module ...', color='orange')
            self._executor(
                ['go', 'mod', 'init', test_config.module_name],
                cwd=build_dir,
                verbose=verbose,
            )
            print_info('Tidy GO module ...', color='orange')
            self._executor(
                ['go', 'mod', 'tidy'],
                cwd=build_dir,
                verbose=verbose,
            )
        except ExecutionFailedError:
            print_info('ERROR', color='red')
            raise BuildFailedError('can\'t init GO module')

        try:
            print_info(f'Building {test_config.module_name} ...', color='orange')
            self._executor(
                ['go', 'build'],
                cwd=build_dir,
                verbose=verbose,
            )
        except ExecutionFailedError:
            print_info('ERROR', color='red')
            raise BuildFailedError(f'Can\'t build {test_binary}')


    def _clean_build(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            verbose: bool = False,
    ) -> None:
        # self._executor(
        #     ['rm', '-rf', str(build_dir)],
        #     check=False,
        #     verbose=verbose,
        # )
        pass

    def _run_tests(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            sandbox: bool = False,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> float:
        stdin = None
        try:
            print_info(f'Testing {test_config.module_name}...', color='orange')
            # args = test_config.args
            args = []
            # if test_config.input_file != '':
            #     stdin = open(build_dir / test_config.input_file, 'r')
            self._executor(
                ['go', 'test', *args],
                cwd=build_dir,
                verbose=verbose,
                # capture_output=test_config.capture_output,
                timeout=test_config.timeout,
                # stdin=stdin
            )
            if test_config.is_crash_me:
                print_info('ERROR', color='red')
                raise TestsFailedError('Program has not crashed')
            print_info('OK', color='green')
        except TimeoutExpiredError:
            print_info('ERROR', color='red')
            message = f'Your solution exceeded time limit: {test_config.timeout} seconds'
            raise TestsFailedError(message)
        except ExecutionFailedError:
            if not test_config.is_crash_me:
                print_info('ERROR', color='red')
                raise TestsFailedError("Test failed (wrong answer or sanitizer error)")
        finally:
            if stdin is not None:
                stdin.close()

        if test_config.is_crash_me:
            print_info('Program has crashed', color='green')
        else:
            print_info('All tests passed', color='green')
        return 1.



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

from ..exceptions import ExecutionFailedError, TestsFailedError
from ..utils.files import copy_files
from ..utils.print import print_info
from .tester import Tester


class MakeTester(Tester):
    @dataclass
    class TaskTestConfig(Tester.TaskTestConfig):
        test_timeout: int = 60  # seconds

        public_test_files: list[str] = field(default_factory=list)
        private_test_files: list[str] = field(default_factory=list)

    def _gen_build(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            source_dir: Path,
            public_tests_dir: Path | None,
            private_tests_dir: Path | None,
            tests_root_dir: Path,
            sandbox: bool = True,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> None:
        self._executor(
            copy_files,
            source=source_dir,
            target=build_dir,
            ignore_patterns=[],
            verbose=verbose,
        )

        if public_tests_dir is not None:
            self._executor(
                copy_files,
                source=public_tests_dir,
                target=build_dir,
                patterns=test_config.public_test_files,
                verbose=verbose,
            )

        if private_tests_dir is not None:
            self._executor(
                copy_files,
                source=private_tests_dir,
                target=build_dir,
                patterns=test_config.private_test_files,
                verbose=verbose,
            )

    def _clean_build(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            verbose: bool = False,
    ) -> None:
        self._executor(
            ['rm', '-rf', str(build_dir)],
            check=False,
            verbose=verbose,
        )

    def _run_tests(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            sandbox: bool = False,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> float:
        tests_cmd = ['make', '-B']

        tests_err = None
        try:
            print_info('Running tests...', color='orange')
            output = self._executor(
                tests_cmd,
                sandbox=sandbox,
                cwd=str(build_dir),
                timeout=test_config.test_timeout,
                verbose=verbose,
                capture_output=True,
            )
            print_info(output, end='')
            print_info('OK', color='green')
        except ExecutionFailedError as e:
            tests_err = e
            print_info(e.output, end='')
            print_info('ERROR', color='red')

        if tests_err is not None:
            raise TestsFailedError('Tests error', output=tests_err.output) from tests_err

        return 1.



from __future__ import annotations

import json
import tempfile
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

from ..course import CourseConfig
from ..exceptions import RunFailedError, TaskTesterTestConfigException, TesterNotImplemented
from ..executors.sandbox import Sandbox
from ..utils.print import print_info


def _create_external_tester(tester_path: Path, dry_run: bool, cleanup: bool) -> Tester:
    globls: Dict[str, Any] = {}
    with open(tester_path) as f:
        tester_code = compile(f.read(), tester_path.absolute(), 'exec')
        exec(tester_code, globls)
    tester_cls = globls.get('CustomTester')
    if tester_cls is None:
        raise TesterNotImplemented(f'class CustomTester not found in file {tester_path}')
    if not issubclass(tester_cls, Tester):
        raise TesterNotImplemented(f'class CustomTester in {tester_path} is not inherited from testers.Tester')
    return tester_cls(dry_run=dry_run, cleanup=cleanup)


class Tester:
    """Entrypoint to testing system
    Tester holds the course object and manage testing of single tasks,
    as well as manage Tasks-Folders mapping
    (Task names for tests should be provided)
    """

    SOURCE_FILES_EXTENSIONS: list[str] = []
    __test__ = False  # to disable pytest detecting it as Test class

    @dataclass
    class TaskTestConfig:
        """Task Tests Config
        Configure how task will copy files, check, execute and so on
        """

        @classmethod
        def from_json(
                cls,
                test_config: Path,
        ) -> 'Tester.TaskTestConfig':
            """
            Create TaskTestConfig from json config
            @param test_config: Path to the config
            @return: TaskTestConfig object
            """
            try:
                task_config_path = test_config
                if task_config_path.exists():
                    with open(task_config_path) as f:
                        raw_config = json.load(f)
                    if not isinstance(raw_config, dict):
                        raise TaskTesterTestConfigException(f'Got <{type(raw_config).__name__}> instead of <dict>')
                else:
                    raw_config = {}
            except (json.JSONDecodeError, TypeError) as e:
                raise TaskTesterTestConfigException(f'Got invalid Test Config <{test_config}>') from e

            # Go throughout config fields and pop it from json if any
            config_kwargs: dict[str, Any] = {}
            for config_field in cls.__annotations__:
                if (field_value := raw_config.pop(config_field, None)) is not None:
                    config_kwargs[config_field] = field_value

            if raw_config:
                bad_keys = ','.join(raw_config.keys())
                raise TaskTesterTestConfigException(f'Test Config {test_config} has unknown key(s) <{bad_keys}>')

            return cls(**config_kwargs)

    def __init__(
            self,
            cleanup: bool = True,
            dry_run: bool = False,
    ):
        self.cleanup = cleanup
        self.dry_run = dry_run
        self._executor = Sandbox(dry_run=dry_run)

    @classmethod
    def create(
            cls,
            root: Path,
            course_config: CourseConfig,
            cleanup: bool = True,
            dry_run: bool = False,
    ) -> 'Tester':
        """
        Main creation entrypoint to Tester
        Create one of existed Testers (python, cpp, etc.)
        @param system: Type of the testing system
        @param cleanup: Perform cleanup after testing
        @param dry_run: Setup dry run mode (really executes nothing)
        @return: Configured Tester object (python, cpp, etc.)
        """
        system = course_config.system
        if system == 'python':
            from . import python
            return python.PythonTester(cleanup=cleanup, dry_run=dry_run)
        elif system == 'make':
            from . import make
            return make.MakeTester(cleanup=cleanup, dry_run=dry_run)
        elif system == 'cpp':
            from . import cpp
            return cpp.CppTester(cleanup=cleanup, dry_run=dry_run)
        elif system == 'go':
            from . import go
            return go.GoTester(cleanup=cleanup, dry_run=dry_run)
        elif system == 'external':
            path = course_config.tester_path
            if path is None:
                raise TesterNotImplemented('tester_path is not specified in course config')
            return _create_external_tester(root / path, cleanup=cleanup, dry_run=dry_run)
        else:
            raise TesterNotImplemented(f'Tester for <{system}> are not supported right now')

    @abstractmethod
    def _gen_build(
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            source_dir: Path,
            public_tests_dir: Path | None,
            private_tests_dir: Path | None,
            tests_root_dir: Path,
            sandbox: bool = True,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> None:  # pragma: nocover
        """
        Copy all files for testing and build the program (if necessary)
        @param test_config: Test config to pass into each stage
        @param build_dir: Directory to copy files into and build there
        @param source_dir: Solution source code directory
        @param public_tests_dir: Directory to copy public tests from
        @param private_tests_dir: Directory to copy private tests from
        @param sandbox: Wrap all student's code to sandbox; @see Executor.sandbox
        @param verbose: Verbose output (can exhibit private tests information)
        @param normalize_output: Normalize all stages output to stderr
        @return: None
        """
        pass

    @abstractmethod
    def _clean_build(
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            verbose: bool = False,
    ) -> None:  # pragma: nocover
        """
        Clean build directory after testing
        @param test_config: Test config to pass into each stage
        @param build_dir: Build directory to clean up
        @param verbose: Verbose output (can exhibit private tests information)
        @return: None
        """
        pass

    @abstractmethod
    def _run_tests(
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            sandbox: bool = False,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> float:  # pragma: nocover
        """
        Run tests for already built task and return solution score
        @param test_config: Test config to pass into each stage
        @param build_dir: Directory with task ready for testing
        @param sandbox: Wrap all student's code to sandbox; @see Executor.sandbox
        @param verbose: Verbose output (can exhibit private tests information)
        @param normalize_output: Normalize all stages output to stderr
        @return: Percentage of the final score
        """
        pass

    def test_task(
            self,
            source_dir: Path,
            config_dir: Path,
            public_tests_dir: Path | None,
            private_tests_dir: Path | None,
            tests_root_dir: Path,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> float:
        """ Inner function to test the task (Folders already specified)
        Perform the following actions:
        * _gen_build: copy source and test files, check forbidden regxp, build and install if necessary
        * _run_tests: run testing and linting
        * _clean_build: cleanup if necessary
        @param source_dir: Solution dir (student's solution or authors' solution)
        @param config_dir: Directory with task config
        @param public_tests_dir: Directory to copy public tests from
        @param private_tests_dir: Directory to copy private tests from
        @param verbose: Verbose output (can exhibit private tests information)
        @param normalize_output: Normalize all stages output to stderr
        @raise RunFailedError: on any build/test error
        @return: Percentage of the final score
        """
        # Read test config
        test_config = self.TaskTestConfig.from_json(config_dir / '.tester.json')

        # Create build dir as tmp dir
        build_dir = Path(tempfile.mkdtemp())
        build_dir.chmod(0o777)  # Set mode for build directory (for code generation and so on)

        try:
            self._gen_build(
                test_config,
                build_dir,
                source_dir,
                public_tests_dir,
                private_tests_dir,
                tests_root_dir,
                sandbox=True,
                verbose=verbose,
                normalize_output=normalize_output,
            )

            # Do not disable sandbox (otherwise it will not clear environ,
            # so environ-related issues may be missed, such as empty locale)
            score_percentage = self._run_tests(
                test_config,
                build_dir,
                sandbox=True,
                verbose=verbose,
                normalize_output=normalize_output
            )
        except RunFailedError as e:
            print_info('\nOoops... Something went wrong: ' + e.msg + (e.output or ''), color='red')
            raise e
        finally:
            if self.cleanup:
                self._clean_build(
                    test_config,
                    build_dir,
                    verbose=verbose
                )
            else:
                print_info(f'Keeping build directory: {build_dir}')

        return score_percentage



from __future__ import annotations

from pydantic import AnyUrl

from .base import PluginABC, PluginOutput


class CheckGitlabMergeRequestPlugin(PluginABC):
    """Plugin for checking gitlab merge request."""

    name = "check_gitlab_merge_request"

    class Args(PluginABC.Args):
        token: str
        task_dir: str
        repo_url: AnyUrl
        requre_approval: bool = False
        search_for_score: bool = False

    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:  # type: ignore[override]
        # TODO: implement
        assert NotImplementedError()

        return PluginOutput(
            output="",
        )


class CollectScoreGitlabMergeRequestPlugin(PluginABC):
    """Plugin for collecting score from gitlab merge request."""

    name = "collect_score_gitlab_merge_request"

    class Args(PluginABC.Args):
        token: str
        task_dir: str
        repo_url: AnyUrl
        requre_approval: bool = False
        search_for_score: bool = False

    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:  # type: ignore[override]
        # TODO: implement
        assert NotImplementedError()

        # TODO: implement arithmetical operations in score comment
        # e.g 0.3 + 2*0.7 - 0.2
        # TODO: auto detect percentage or 0-1 score
        # e.g. valid:
        # 0.3 + 2*0.7 - 0.2 = 0.8
        # 30% + 70% - 20% = 80% (return as 0.8)
        # 30 + 70 - 20 = 80 (return as 0.8)

        return PluginOutput(
            output="",
            percentage=1.0,
        )



"""
Classes for Course, Groups and Tasks
They interact ONLY with deadlines, do not test anything, know nothing about physical folders
"""
from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from pathlib import Path

import yaml

from ..exceptions import BadConfig, BadGroupConfig, BadTaskConfig


RESERVED_TASK_NAMES = [
    'task', 'tasks', 'solution', 'solutions', 'test', 'tests', 'lecture', 'lectures', 'template',
    'templates', 'private', 'privates', 'public', 'publics', 'review', 'reviews',
]


@dataclass
class Task:
    group: 'Group'
    name: str
    full_name: str = field(init=False)

    max_score: int
    enabled: bool = True
    scoring_func: str = 'max'
    review: bool = False
    marked: bool = False

    def __post_init__(self) -> None:
        self.full_name = self.group.name + '/' + self.name

        assert self.name not in RESERVED_TASK_NAMES, f'Can not use {self.name} as task name as it is reserved'

    @property
    def is_enabled(self) -> bool:
        return self.enabled and self.group.is_enabled

    @property
    def is_started(self) -> bool:
        return self.is_enabled and self.group.is_started

    @property
    def is_ended(self) -> bool:
        return self.is_enabled and self.group.is_ended

    def get_task_deadline_percentage(
            self,
            submit_time: datetime | None = None,
            extra_time: timedelta | None = None,
    ) -> float:
        return self.group.get_deadline_percentage(submit_time, extra_time)

    def get_is_overdue_first(
            self,
            submit_time: datetime | None = None,
            extra_time: timedelta | None = None,
    ) -> bool:
        return self.group.get_is_overdue_first(submit_time=submit_time, extra_time=extra_time)

    def get_is_overdue_second(
            self,
            submit_time: datetime | None = None,
            extra_time: timedelta | None = None,
    ) -> bool:
        return self.group.get_is_overdue_second(submit_time=submit_time, extra_time=extra_time)


@dataclass
class Group:
    name: str

    start: datetime
    deadline: datetime
    second_deadline: datetime

    enabled: bool = True
    marked: bool = False

    tasks: list[Task] = field(default_factory=list)

    @property
    def max_score(self) -> int:
        return sum([task.max_score for task in self.tasks])

    @property
    def is_enabled(self) -> bool:
        return self.enabled

    @property
    def is_started(self) -> bool:
        return self.is_enabled and self.start < datetime.now()  # TODO: check timezone

    @property
    def is_ended(self) -> bool:
        return self.is_enabled and self.second_deadline < datetime.now()  # TODO: check timezone

    def get_deadline_percentage(
            self,
            submit_time: datetime | None = None,
            extra_time: timedelta | None = None,
    ) -> float:
        extra_time = extra_time or timedelta()
        submit_time = submit_time or datetime.now()  # TODO: check timezone
        if self.second_deadline == self.deadline:
            return 1. if submit_time < self.second_deadline + extra_time else 0.

        deadlines_timedelta = self.second_deadline - self.deadline
        overdue_timedelta = self.second_deadline + extra_time - submit_time
        percentage = overdue_timedelta / deadlines_timedelta
        return max(0., min(percentage, 1.))

    def get_is_overdue_first(
            self,
            submit_time: datetime | None = None,
            extra_time: timedelta | None = None,
    ) -> bool:
        return self.get_deadline_percentage(submit_time, extra_time) < 1.

    def get_is_overdue_second(
            self,
            submit_time: datetime | None = None,
            extra_time: timedelta | None = None,
    ) -> bool:
        return self.get_deadline_percentage(submit_time, extra_time) == 0.


class CourseSchedule:
    def __init__(
            self,
            deadlines_config: Path,
    ):
        try:
            with open(deadlines_config) as config_file:
                deadlines = yaml.safe_load(config_file)
        except (yaml.YAMLError, FileNotFoundError) as e:
            raise BadConfig(f'Unable to load deadlines config file <{deadlines_config}>') from e

        if not deadlines:
            raise BadConfig(f'Empty config file <{deadlines_config}>')

        self.groups: OrderedDict[str, Group] = OrderedDict()
        self.tasks: OrderedDict[str, Task] = OrderedDict()

        for group_config in deadlines:
            group_name = None
            try:
                group_name = str(group_config.get('group'))
                group_enabled = bool(group_config.get('enabled', True))

                group_start = datetime.strptime(group_config.get('start'), '%d-%m-%Y %H:%M')
                group_deadline = datetime.strptime(group_config.get('deadline'), '%d-%m-%Y %H:%M')
                if second_deadline := group_config.get('second_deadline', None):
                    group_second_deadline = datetime.strptime(second_deadline, '%d-%m-%Y %H:%M')
                else:
                    group_second_deadline = group_deadline

                group_marked = bool(group_config.get('marked', False))
            except (KeyError, TypeError, ValueError, AttributeError) as e:
                raise BadGroupConfig(f'Group {group_name} has bad config') from e

            group = Group(
                name=group_name,
                enabled=group_enabled,
                start=group_start,
                deadline=group_deadline,
                second_deadline=group_second_deadline,
                marked=group_marked,
            )

            for task_config in group_config.get('tasks', []):
                task_name = None
                try:
                    task_name = task_config['task']
                    task_score = int(task_config['score'])
                    task_enabled = task_config.get('enabled', True)
                    task_scoring_func = task_config.get('scoring_func', 'max')
                    task_is_review = task_config.get('review', False)
                    task_marked = task_config.get('marked', False) or group_marked
                except (KeyError, TypeError, ValueError, AttributeError) as e:
                    raise BadTaskConfig(f'Task {task_name} has bad config') from e

                task = Task(
                    group=group,
                    name=task_name,
                    max_score=task_score,
                    enabled=task_enabled,
                    scoring_func=task_scoring_func,
                    review=task_is_review,
                    marked=task_marked,
                )

                if task_name in self.tasks:
                    raise BadTaskConfig(f'Unique violation error: task {task_name} already exists')

                self.tasks[task_name] = task
                group.tasks.append(task)

            if group_name in self.groups:
                raise BadGroupConfig(f'Unique violation error: group {group_name} already exists')

            self.groups[group_name] = group

    def get_tasks(
            self,
            *,
            enabled: bool | None = None,
            started: bool | None = None,
            ended: bool | None = None,
    ) -> list[Task]:
        tasks: list[Task] = [task for task_name, task in self.tasks.items()]

        if enabled is not None:
            tasks = [task for task in tasks if (task.is_enabled and task.group.is_enabled) == enabled]
        if started is not None:
            tasks = [task for task in tasks if task.group.is_started == started]
        if ended is not None:
            tasks = [task for task in tasks if task.group.is_ended == ended]

        return tasks

    def get_groups(
            self,
            *,
            enabled: bool | None = None,
            started: bool | None = None,
            ended: bool | None = None,
    ) -> list[Group]:
        groups: list[Group] = [group for group_name, group in self.groups.items()]

        if enabled is not None:
            groups = [group for group in groups if group.is_enabled == enabled]
        if started is not None:
            groups = [group for group in groups if group.is_started == started]
        if ended is not None:
            groups = [group for group in groups if group.is_ended == ended]

        return groups



from __future__ import annotations

import re
from dataclasses import InitVar, dataclass, field
from pathlib import Path

from ..exceptions import BuildFailedError, ExecutionFailedError, RunFailedError, StylecheckFailedError, TestsFailedError
from ..utils.files import check_folder_contains_regexp, copy_files
from ..utils.print import print_info
from .tester import Tester


IGNORE_FILE_PATTERNS = ['*.md', 'build', '__pycache__', '.pytest_cache', '.mypy_cache', '.tester.json']
COVER_IGNORE_FILES = ['setup.py']


class PythonTester(Tester):

    SOURCE_FILES_EXTENSIONS: list[str] = ['.py']

    @dataclass
    class TaskTestConfig(Tester.TaskTestConfig):
        partially_scored: bool = False
        verbose_tests_output: bool = False
        module_test: bool = False
        build_wheel: bool = False
        run_mypy: bool = True

        forbidden_regexp: list[re.Pattern[str]] = field(default_factory=list)

        public_test_files: list[str] = field(default_factory=list)
        private_test_files: list[str] = field(default_factory=list)

        test_timeout: int = 60  # seconds
        coverage: bool | int = False

        # Created on init
        test_files: list[str] = field(init=False, default_factory=list)
        # Init only
        explicit_public_tests: InitVar[list[str]] = None
        explicit_private_tests: InitVar[list[str]] = None

        def __post_init__(
                self,
                explicit_public_tests: list[str] | None,
                explicit_private_tests: list[str] | None,
        ) -> None:
            self.forbidden_regexp += [r'exit\(0\)']  # type: ignore
            for regexp in self.forbidden_regexp:
                re.compile(regexp)

            self.public_test_files = ['test_public.py'] + (explicit_public_tests or [])
            self.private_test_files = ['test_private.py'] + (explicit_private_tests or [])
            self.test_files = self.public_test_files + self.private_test_files

    def _gen_build(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            source_dir: Path,
            public_tests_dir: Path | None,
            private_tests_dir: Path | None,
            tests_root_dir: Path,
            sandbox: bool = True,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> None:
        # Copy submitted code (ignore tests)
        self._executor(
            copy_files,
            source=source_dir,
            target=build_dir,
            ignore_patterns=test_config.test_files + IGNORE_FILE_PATTERNS,
            verbose=verbose,
        )

        # Check submitted code using forbidden regexp
        self._executor(
            check_folder_contains_regexp,
            folder=build_dir,
            extensions=self.SOURCE_FILES_EXTENSIONS,
            regexps=test_config.forbidden_regexp,
            raise_on_found=True,
            verbose=verbose,
        )

        # Install submitted code as module if needed
        if test_config.module_test:
            # assert setup files exists
            setup_files = {i.name for i in build_dir.glob(r'setup.*')} | \
                          {i.name for i in build_dir.glob(r'pyproject.*')}
            if 'setup.py' not in setup_files and 'setup.cfg' not in setup_files and 'pyproject.toml' not in setup_files:
                raise BuildFailedError(
                    'This task is in editable `module` mode. You have to provide pyproject.toml/setup.cfg/setup.py file'
                )
            if 'setup.py' not in setup_files:
                raise BuildFailedError('This task is in editable `module` mode. You have to provide setup.py file')

            if test_config.build_wheel:
                task_build_dir_dist = build_dir / 'dist'
                output = self._executor(
                    ['pip3', 'wheel', '--wheel-dir', str(task_build_dir_dist), str(build_dir)],
                    verbose=verbose,
                    env_sandbox=sandbox,
                    capture_output=normalize_output,
                )
                if normalize_output:
                    print_info(output or '', end='')

                output = self._executor(
                    ['pip3', 'install', '--prefer-binary', '--force-reinstall', '--find-links',
                     str(task_build_dir_dist), str(build_dir)],
                    verbose=verbose,
                    env_sandbox=sandbox,
                    capture_output=normalize_output,
                )
                if normalize_output:
                    print_info(output or '', end='')

                if (build_dir / 'build').exists():
                    output = self._executor(
                        ['rm', '-rf', str(build_dir / 'build')],
                        verbose=verbose,
                        env_sandbox=sandbox,
                        capture_output=normalize_output,
                    )
                    if normalize_output:
                        print_info(output or '', end='')
            else:
                output = self._executor(
                    ['pip3', 'install', '-e', str(build_dir), '--force'],
                    verbose=verbose,
                    env_sandbox=sandbox,
                    capture_output=normalize_output,
                )
                if normalize_output:
                    print_info(output or '', end='')

        # Copy public test files
        if public_tests_dir is not None:
            self._executor(
                copy_files,
                source=public_tests_dir,
                target=build_dir,
                patterns=test_config.public_test_files,
                verbose=verbose,
            )

        # Copy private test files
        if private_tests_dir is not None:
            self._executor(
                copy_files,
                source=private_tests_dir,
                target=build_dir,
                patterns=test_config.private_test_files,
                verbose=verbose,
            )

    def _clean_build(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            verbose: bool = False,
    ) -> None:
        self._executor(
            ['rm', '-rf', str(build_dir)],
            check=False,
            verbose=verbose,
        )

    @staticmethod
    def _parse_summary_score(
            output: str,
    ) -> float:
        score = 0.0
        for line in output.splitlines():
            if 'Summary score percentage is: ' in line:
                score += float(line.strip().split('Summary score percentage is: ')[1])
                break
        return score

    def _run_tests(  # type: ignore[override]
            self,
            test_config: TaskTestConfig,
            build_dir: Path,
            sandbox: bool = False,
            verbose: bool = False,
            normalize_output: bool = False,
    ) -> float:
        # TODO: replace with preserved setup.cfg
        codestyle_cmd = [
            'flake8',
            '--exclude', ','.join(test_config.private_test_files),
            '--max-line-length', '120',
            str(build_dir)
        ]
        # codestyle_cmd = [
        #     'ruff',
        #     '--exclude', ','.join(test_config.private_test_files),
        #     '--line-length', '120',
        #     '--no-fix',
        #     str(build_dir)
        # ]
        mypy_cmd = [
            'mypy',
            '--no-incremental',
            '--cache-dir', '/dev/null',
            '--ignore-missing-imports',
            '--disallow-untyped-defs',
            '--disallow-incomplete-defs',
            '--disallow-subclassing-any',
            '--disallow-any-generics',
            '--no-implicit-optional',
            '--warn-redundant-casts',
            '--warn-unused-ignores',
            '--warn-unreachable',
            '--allow-untyped-decorators',
            str(build_dir)
        ]
        tests_collection_cmd = [
            'pytest',
            '-p', 'no:cacheprovider',
            '-p', 'no:requests_mock',
            '-p', 'no:cov',
            '-p', 'no:mock',
            '-p', 'no:socket',
            '-qq',
            '--collect-only',
            str(build_dir)
        ]
        tests_cmd = [
            'pytest',
            '-p', 'no:cacheprovider',
            '-p', 'no:requests_mock',
            '-p', 'no:timeout',
            '-p', 'no:socket',
            # '--timeout=60',
            str(build_dir)
        ]
        if not verbose:
            tests_cmd += ['--no-header']
        if not verbose and not test_config.verbose_tests_output:
            tests_cmd += ['--tb=no']
        # if test_config.partially_scored:
        #     tests_cmd += ['-s']
        if test_config.coverage:
            tests_cmd += ['--cov-report', 'term-missing']

            # exclude test files
            dirs_to_cover = {
                i.relative_to(build_dir) for i in build_dir.iterdir()
                if i.suffix in ['', '.py'] and i.name not in test_config.test_files and i.name not in COVER_IGNORE_FILES
            }
            if dirs_to_cover:
                for _dir in dirs_to_cover:
                    tests_cmd += ['--cov', str(_dir).replace(r'.', r'\.')]
            else:
                tests_cmd += ['--cov', str(build_dir)]

            # tests_cmd += ['--cov-config', '.coveragerc']
            if test_config.coverage is not True:
                tests_cmd += ['--cov-fail-under', str(test_config.coverage)]
        else:
            tests_cmd += ['-p', 'no:cov']

        # Check style
        styles_err = None
        try:
            print_info('Running codestyle checks...', color='orange')
            output = self._executor(
                codestyle_cmd,
                sandbox=sandbox,
                cwd=str(build_dir),
                verbose=verbose,
                capture_output=normalize_output,
            )
            if normalize_output:
                print_info(output or '', end='')
            print_info('[No issues]')
            print_info('OK', color='green')
        except ExecutionFailedError as e:
            # Always reraise for style checks
            styles_err = e

            if normalize_output:
                print_info(e.output, end='')
                e.output = ''
                output = ''
            print_info('ERROR', color='red')

        # Check typing
        typing_err = None
        try:
            if test_config.run_mypy:
                print_info('Running mypy checks...', color='orange')
                output = self._executor(
                    mypy_cmd,
                    sandbox=sandbox,
                    cwd=str(build_dir.parent),  # mypy didn't work from cwd
                    verbose=verbose,
                    capture_output=normalize_output,
                )
                if normalize_output:
                    print_info(output, end='')
                print_info('OK', color='green')
            else:
                print_info('Type check is skipped for this task!', color='orange')
        except ExecutionFailedError as e:
            # Always reraise for typing checks
            typing_err = e

            if normalize_output:
                print_info(e.output, end='')
                e.output = ''
                output = ''
            print_info('ERROR', color='red')

        # Check import and tests collecting
        import_err = None
        try:
            print_info('Collecting tests...', color='orange')
            output = self._executor(
                tests_collection_cmd,
                sandbox=sandbox,
                cwd=str(build_dir),
                verbose=verbose,
                capture_output=normalize_output,
            )
            if normalize_output:
                print_info(output, end='')
                output = ''
            print_info('OK', color='green')
        except ExecutionFailedError as e:
            # Always reraise for import checks
            import_err = e

            if normalize_output:
                print_info(e.output, end='')
                e.output = ''
                output = ''
            print_info('ERROR', color='red')

        # Check tests
        tests_err = None
        tests_output = ''
        try:
            print_info('Running tests...', color='orange')
            output = self._executor(
                tests_cmd,
                sandbox=sandbox,
                cwd=str(build_dir),
                timeout=test_config.test_timeout,
                verbose=verbose,
                capture_output=test_config.partially_scored or normalize_output,
            )
            if normalize_output or test_config.partially_scored:
                print_info(output, end='')
            print_info('OK', color='green')
        except ExecutionFailedError as e:
            tests_err = e
            tests_output = e.output or ''

            if normalize_output or test_config.partially_scored:
                print_info(e.output, end='')
                e.output = ''
                output = ''

            if test_config.partially_scored:
                print_info('ERROR? (Some tests failed, but this is partially_scored task)', color='orange')
            else:
                print_info('ERROR', color='red')

        if import_err is not None:
            raise RunFailedError('Import error', output=import_err.output) from import_err

        if tests_err is not None and not test_config.partially_scored:  # Reraise only if all tests should pass
            raise TestsFailedError('Public or private tests error', output=tests_err.output) from tests_err

        if styles_err is not None:
            raise StylecheckFailedError('Style error', output=styles_err.output) from styles_err

        if typing_err is not None:
            raise StylecheckFailedError('Typing error', output=typing_err.output) from typing_err

        if test_config.partially_scored:
            tests_output = tests_output or ''  # for mypy only
            return self._parse_summary_score(tests_output)
        else:
            return 1.
